#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-07-11
# @Filename: notifier.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Any, Callable, Coroutine, List, TypeVar

from .message import Message


if TYPE_CHECKING:
    from .bus import BusABC

__all__ = ["Notifier"]


Listener_co = Callable[..., Coroutine[Message, Any, Any]]
Bus_co = TypeVar("Bus_co", bound="BusABC")


class Notifier:
    """Notifier class to report bus messages to multiple listeners."""

    def __init__(self, listeners: List[Listener_co] = [], buses: List[Bus_co] = []):

        self.loop = asyncio.get_running_loop()

        self.listeners = listeners

        self.tasks: list[asyncio.Task] = []

        self.buses: List[BusABC] = []
        for bus in buses:
            self.add_bus(bus)

    def stop(self):
        """Stops the notifier."""

        for task in self.tasks:
            task.cancel()

        self.buses = []

    def add_listener(self, callback: Listener_co):
        """Adds a listener."""

        self.listeners.append(callback)

    def add_bus(self, bus):
        """Adds a bus to monitor."""

        self.buses.append(bus)
        self.tasks.append(asyncio.create_task(self._monitor_bus(bus)))

    def remove_notifier(self, callback: Listener_co):
        """Removes a listener."""

        if callback in self.listeners:
            self.listeners.remove(callback)

    async def _monitor_bus(self, bus: BusABC):
        """Monitors buses and calls the listeners when a message is received."""

        while True:
            msg = await bus.get()
            if msg is not None:
                for listener in self.listeners:
                    asyncio.create_task(listener(msg))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-13
# @Filename: generate_chiller_yaml.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import sys

import pandas
import yaml


BASE = """
address: 10.25.1.162
port: 1111
modules:
    CHILLER:
        mode: holding_register
        channels: -1
        description: APO chiller status variables
        devices: {}
"""


NAME_CONV = {
    "STATUS_FLUID_FLOW_SV": "STATUS_FLUID_FLOW",
    "USER_FLOW_SP_GPM_SV": "FLOW_USER_SETPOINT",
    "STATUS_DISPLAY_VALUE_SV": "DISPLAY_VALUE",
    "USER_SETPOINT_SV": "TEMPERATURE_USER_SETPOINT",
    "STATUS_AMBIENT_AIR_SV": "STATUS_AMBIENT_AIR",
}


def generate_chiller_yaml(variables_files: str):
    """Generates a YAML file for Drift with all the chiller variables from the CSV."""

    variables = pandas.read_csv(variables_files)

    data = yaml.load(BASE, yaml.SafeLoader)
    devices = {}

    for _, row in variables.iterrows():
        address = row.Address - 1

        name = row.Name.upper()
        if name in NAME_CONV:
            name = NAME_CONV[name]

        devices[name] = {
            "address": address,
            "units": row.Unit if isinstance(row.Unit, str) else "",
            "category": "chiller",
            "description": row.Description,
        }

        if row.Scale != 1:
            devices[name].update(
                {
                    "adaptor": "proportional",
                    "adaptor_extra_params": [0.1],
                }
            )

    data["modules"]["CHILLER"]["devices"] = devices

    with open("chiller.yaml", "w") as f:
        yaml.dump(data, f, yaml.SafeDumper)


if __name__ == "__main__":
    generate_chiller_yaml(sys.argv[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-05-12
# @Filename: ieb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import math
import os
import warnings

from typing import Any, Dict

from drift import Drift, DriftError

from jaeger import config
from jaeger.exceptions import JaegerUserWarning


__all__ = ["IEB", "FVC", "_get_category_data"]


async def _get_category_data(command, category) -> list:
    """Returns data about a device category."""

    ieb = command.actor.fps.ieb
    schema = command.actor.model.schema

    items = schema["properties"][category]["items"]
    measured = []

    async with ieb:
        for item in items:
            name = item["title"]
            type_ = item["type"]
            device = ieb.get_device(name)
            value = (await device.read(connect=False))[0]
            if type_ == "boolean" and device.__type__ == "relay":
                value = True if value == "closed" else False
            elif type_ == "integer":
                value = int(value)
            elif type_ == "number":
                if "multipleOf" in item:
                    precision = int(-math.log10(item["multipleOf"]))
                else:
                    precision = 3
                value = round(value, precision)
            measured.append(value)

    return measured


class IEB(Drift):
    """Thing wrapper around a :class:`~drift.drift.Drift` class.

    Allows additional features such as disabling the interface.

    """

    MAX_RETRIES: int = 5

    def __init__(self, *args, **kwargs):

        self.disabled = False

        super().__init__(*args, **kwargs)

        self._categories = None
        self._n_failures: int = 0

    @classmethod
    def create(cls, path=None):
        """Creates an `.IEB` instance with the default configuration."""

        default_ieb_path = path or config["files"]["ieb_config"]

        default_ieb_path = os.path.expanduser(os.path.expandvars(default_ieb_path))
        if not os.path.isabs(default_ieb_path):
            default_ieb_path = os.path.join(os.path.dirname(__file__), default_ieb_path)

        return cls.from_config(default_ieb_path)

    def get_categories(self):
        """Returns a list of categories."""

        if self._categories is None:
            categories = [
                device.category
                for module in self.modules
                for device in self.modules[module].devices.values()
            ]
            self._categories = set(categories)
            self._categories.discard(None)

        return self._categories

    async def __aenter__(self):

        if self.disabled:
            raise DriftError("IEB is disabled.")

        n_retries = 0
        while True:
            try:
                await Drift.__aenter__(self)
                break
            except DriftError:
                n_retries += 1
                if n_retries >= 5:
                    raise DriftError("Failed connecting to the IEB.")

    async def __aexit__(self, *args):

        await Drift.__aexit__(self, *args)

    def enable(self):
        """Re-enables the IEB instance."""

        self._n_failures = 0
        self.disabled = False

    async def get_status(self) -> Dict[str, Any]:
        """Returns the status of the IEB components."""

        status = {}
        for category in self.get_categories():
            data = await self.read_category(category)
            for device in data:
                if self.get_device(device).__type__ == "relay":
                    value = False if data[device][0] == "open" else True
                else:
                    value = data[device][0]
                status[device] = value

        return status


class FVC(IEB):
    """Connects to the FVC IEB."""

    @classmethod
    def create(cls, path=None):
        """Creates an `.FVC` instance with the default configuration."""

        default_ieb_path = config["fvc"]["config"]

        return super().create(default_ieb_path)


class Chiller(IEB):
    """Connects to the chiller Modbus PLC."""

    @classmethod
    def create(cls, path=None):
        """Creates a `.Chiller` instance with the default configuration."""

        config_file = config["chiller"].get("config", None)
        if config_file is None:
            warnings.warn("Chiller configuration missing.", JaegerUserWarning)
            return None

        return super().create(config_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-11
# @Filename: maskbits.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import enum


__all__ = [
    "Maskbit",
    "PositionerStatus",
    "CommandStatus",
    "ResponseCode",
    "BootloaderStatus",
    "FPSStatus",
]


class Maskbit(enum.IntFlag):
    """A maskbit enumeration. Intended for subclassing."""

    __version__ = None

    def __str__(self):
        members, _ = enum._decompose(self.__class__, self._value_)  # type: ignore
        return "|".join([str(m._name_ or m._value_) for m in members])

    @property
    def version(self):
        """The version of the flags."""

        return self.__version__

    @property
    def active_bits(self):
        """Returns a list of flags that match the value."""

        return [bit for bit in self.__class__ if bit.value & self.value]  # type: ignore


class CommandStatus(Maskbit):
    """Maskbits for command status."""

    DONE = enum.auto()
    CANCELLED = enum.auto()
    FAILED = enum.auto()
    READY = enum.auto()
    RUNNING = enum.auto()
    TIMEDOUT = enum.auto()

    @property
    def is_done(self):
        """Returns True if the command is done (completed or failed)."""

        return True if (self in [self.DONE, self.TIMEDOUT] or self.failed) else False

    @property
    def is_running(self):
        """Returns True if the command is running."""

        return True if self == CommandStatus.RUNNING else False

    @property
    def failed(self):
        """Returns True if the command failed (or got cancelled)."""

        failed_states = self.CANCELLED | self.FAILED
        return True if self in failed_states else False

    @property
    def timed_out(self):
        """Returns True if the command timed out."""

        return True if self.TIMEDOUT in self else False


class PositionerStatusV4_1(Maskbit):
    """Maskbits for positioner status (firmware >=04.01.00)."""

    __version__ = "4.1"

    SYSTEM_INITIALIZED = 0x0000000000000001
    CONFIG_CHANGED = 0x0000000000000002
    BSETTINGS_CHANGED = 0x0000000000000004
    DATA_STREAMING = 0x0000000000000008
    RECEIVING_TRAJECTORY = 0x0000000000000010
    TRAJECTORY_ALPHA_RECEIVED = 0x0000000000000020
    TRAJECTORY_BETA_RECEIVED = 0x0000000000000040
    LOW_POWER_AFTER_MOVE = 0x0000000000000080
    DISPLACEMENT_COMPLETED = 0x0000000000000100
    DISPLACEMENT_COMPLETED_ALPHA = 0x0000000000000200
    DISPLACEMENT_COMPLETED_BETA = 0x0000000000000400
    COLLISION_ALPHA = 0x0000000000000800
    COLLISION_BETA = 0x0000000000001000
    CLOSED_LOOP_ALPHA = 0x0000000000002000
    CLOSED_LOOP_BETA = 0x0000000000004000
    PRECISE_POSITIONING_ALPHA = 0x0000000000008000
    PRECISE_POSITIONING_BETA = 0x0000000000010000
    COLLISION_DETECT_ALPHA_DISABLE = 0x0000000000020000
    COLLISION_DETECT_BETA_DISABLE = 0x0000000000040000
    MOTOR_CALIBRATION = 0x0000000000080000
    MOTOR_ALPHA_CALIBRATED = 0x0000000000100000
    MOTOR_BETA_CALIBRATED = 0x0000000000200000
    DATUM_CALIBRATION = 0x0000000000400000
    DATUM_ALPHA_CALIBRATED = 0x0000000000800000
    DATUM_BETA_CALIBRATED = 0x0000000001000000
    DATUM_INITIALIZATION = 0x0000000002000000
    DATUM_ALPHA_INITIALIZED = 0x0000000004000000
    DATUM_BETA_INITIALIZED = 0x0000000008000000
    HALL_ALPHA_DISABLE = 0x0000000010000000
    HALL_BETA_DISABLE = 0x0000000020000000
    COGGING_CALIBRATION = 0x0000000040000000
    COGGING_ALPHA_CALIBRATED = 0x0000000080000000
    COGGING_BETA_CALIBRATED = 0x0000000100000000
    ESTIMATED_POSITION = 0x0000000200000000
    POSITION_RESTORED = 0x0000000400000000
    SWITCH_OFF_AFTER_MOVE = 0x0000000800000000
    CALIBRATION_SAVED = 0x0000001000000000
    PRECISE_MOVE_IN_OPEN_LOOP_ALPHA = 0x0000002000000000
    PRECISE_MOVE_IN_OPEN_LOOP_BETA = 0x0000004000000000
    SWITCH_OFF_HALL_AFTER_MOVE = 0x0000008000000000
    UNKNOWN = 0x0000010000000000

    @property
    def collision(self):
        """Returns `True` if the positioner is collided."""

        return (
            True
            if (
                PositionerStatusV4_1.COLLISION_ALPHA & self
                or PositionerStatusV4_1.COLLISION_BETA & self
            )
            else False
        )

    @property
    def initialised(self):
        """Returns `True` if the positioner is initialised."""

        return True if PositionerStatusV4_1.SYSTEM_INITIALIZED & self else False


class PositionerStatusV4_0(Maskbit):
    """Maskbits for positioner status (firmware <=04.00.04)."""

    __version__ = "4.0"

    SYSTEM_INITIALIZATION = 0x00000001
    RECEIVING_TRAJECTORY = 0x00000100
    TRAJECTORY_ALPHA_RECEIVED = 0x00000200
    TRAJECTORY_BETA_RECEIVED = 0x00000400
    DATUM_INITIALIZATION = 0x00200000
    DATUM_ALPHA_INITIALIZED = 0x00400000
    DATUM_BETA_INITIALIZED = 0x00800000
    DISPLACEMENT_COMPLETED = 0x01000000
    ALPHA_DISPLACEMENT_COMPLETED = 0x02000000
    BETA_DISPLACEMENT_COMPLETED = 0x04000000
    ALPHA_COLLISION = 0x08000000
    BETA_COLLISION = 0x10000000
    DATUM_INITIALIZED = 0x20000000
    ESTIMATED_POSITION = 0x40000000
    POSITION_RESTORED = 0x80000000
    UNKNOWN = 0x100000000

    @property
    def initialised(self):
        """Returns `True` if the positioner is initialised."""

        return True if PositionerStatusV4_0.SYSTEM_INITIALIZATION & self else False

    @property
    def collision(self):
        """Returns `True` if the positioner is collided."""

        return (
            True
            if (
                PositionerStatusV4_0.ALPHA_COLLISION & self
                or PositionerStatusV4_0.BETA_COLLISION & self
            )
            else False
        )


#: Alias to the default positioner status flags.
PositionerStatus = PositionerStatusV4_1


class BootloaderStatus(Maskbit):
    """Maskbits for positioner status when in bootloader mode."""

    BOOTLOADER_INIT = 0x00000001
    BOOTLOADER_TIMEOUT = 0x00000002
    BSETTINGS_CHANGED = 0x00000200
    RECEIVING_NEW_FIRMWARE = 0x00010000
    NEW_FIRMWARE_RECEIVED = 0x01000000
    NEW_FIRMWARE_CHECK_OK = 0x02000000
    NEW_FIRMWARE_CHECK_BAD = 0x04000000
    UNKNOWN = 0x40000000


class ResponseCode(enum.IntFlag):
    """Maskbit for the status of the bootloader.

    - 0: All OK
    - 1: Received parameter is out of range
    - 2: Trajectory not accepted
    - 3: Already in motion, cannot accept command
    - 4: Datum not initialized
    - 5: Incorrect amount of data in packet
    - 6: One of the calibration modes is active: ``MOTOR_CALIBRATION``,
      ``COGGING_CALIBRATION``, ``DATUM_CALIBRATION``, ``DATUM _INITIALIZATION``
    - 7: The motors are not calibrated and therefore the hall sensors can't be used
    - 8: A collision is detected, the flag has to be first cleared with stop trajectory
    - 9: Hall sensors are disabled and can therefore not be used
    - 10: Broadcast command not valid
    - 11: Command not supported by bootloader
    - 12: Invalid command
    - 13: Unknown command
    - 14: Datum not calibrated
    - 15: Halls sensors have been disabled

    """

    COMMAND_ACCEPTED = 0
    VALUE_OUT_OF_RANGE = 1
    INVALID_TRAJECTORY = 2
    ALREADY_IN_MOTION = 3
    DATUM_NOT_INITIALIZED = 4
    INCORRECT_AMOUNT_OF_DATA = 5
    CALIBRATION_MODE_ACTIVE = 6
    MOTOR_NOT_CALIBRATED = 7
    COLLISION_DETECTED = 8
    HALL_SENSOR_DISABLED = 9
    INVALID_BROADCAST_COMMAND = 10
    INVALID_BOOTLOADER_COMMAND = 11
    INVALID_COMMAND = 12
    UNKNOWN_COMMAND = 13
    DATUM_NOT_CALIBRATED = 14
    HALL_SENSORS_DISABLED = 15


class FPSStatus(enum.Flag):
    """Status of the FPS."""

    IDLE = 0x01
    MOVING = 0x02
    COLLIDED = 0x04
    ERRORED = 0x08
    TEMPERATURE_NORMAL = 0x10
    TEMPERATURE_COLD = 0x20
    TEMPERATURE_VERY_COLD = 0x40
    TEMPERATURE_UNKNOWN = 0x80

    TEMPERATURE_BITS = (
        TEMPERATURE_NORMAL
        | TEMPERATURE_COLD
        | TEMPERATURE_VERY_COLD
        | TEMPERATURE_UNKNOWN
    )
    STATUS_BITS = IDLE | ERRORED | MOVING | COLLIDED



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-04
# @Filename: testing.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import JaegerCommandType, jaeger_parser


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = ["testing"]


@jaeger_parser.group()
def testing():
    """Commands for testing. Use with caution."""

    pass


@testing.command(name="fake-configuration")
@click.argument("RA", type=float)
@click.argument("DEC", type=float)
@click.argument("PA", type=float, required=False, default=0.0)
async def disable(
    command: JaegerCommandType,
    fps: FPS,
    ra: float,
    dec: float,
    pa: float = 0.0,
):
    """Outputs the ``configuration_loaded`` keyword for a given field."""

    command.info(
        configuration_loaded=[-999, -999, -999, ra, dec, pa, -999.0, -999.0, "/data"]
    )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-12
# @Filename: database.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from functools import wraps
from glob import glob

from typing import TYPE_CHECKING

import pandas
import peewee
from astropy import table
from astropy.io import fits

from sdssdb.peewee.sdss5db import opsdb, targetdb

from jaeger import config


if TYPE_CHECKING:
    from sdssdb.connection import PeeweeDatabaseConnection

__all__ = [
    "connect_database",
    "load_holes",
    "load_fields",
    "get_designid_from_queue",
    "match_assignment_hash",
]


def connect_database(database: PeeweeDatabaseConnection, force: bool = False):
    """Connects the database if it is not."""

    if database.connected and force is False:
        return True

    database.connect(**config["database"])

    return database.connected


def check_database(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if targetdb.database.connected is False:
            if connect_database(targetdb.database) is False:
                raise RuntimeError("Database is not connected.")
        return f(*args, **kwargs)

    return wrapper


@check_database
def get_designid_from_queue(
    pop: bool = True,
    epoch_delay: bool = False,
) -> tuple[int | None, float | None]:
    """Pops a design from the queue.

    Parameters
    ----------
    pop
        If `True`, pops the next design from the queue. Otherwise just returns the
        design ID.
    epoch_delay
        If `True`, calculates the delta epoch that optimises observing all the
        queued designs with the same hash.

    Returns
    -------
    result
        A tuple with the design ID as the first element and the delta epoch, in
        seconds as the second one. If ``epoch_delay=False``, the delta epoch is
        `None`.


    """

    if pop:
        design = opsdb.Queue.pop()
    else:
        design = (
            opsdb.Queue.select()
            .where(opsdb.Queue.position > 0)
            .order_by(opsdb.Queue.position)
            .first()
        )

    if design is None:
        return (None, None)

    if epoch_delay is False:
        return (design.design_id, None)

    # Get a list of all the entries in the queue that have the
    # same hash as the first one.
    hash = targetdb.Design.get_by_id(design.design_id).assignment_hash.hex

    design_hashes = (
        opsdb.Queue.select(opsdb.Queue.position)
        .join(targetdb.Design, on=(targetdb.Design.design_id == opsdb.Queue.design_id))
        .where(targetdb.Design.assignment_hash == hash)
        .where(opsdb.Queue.position >= (1 if pop is False else -1))
        .order_by(opsdb.Queue.position)
        .tuples()
    )

    # This should not happen, but if the first entry does not have position 1,
    # just leave because something weird happened. Note that the position can be -1
    # if we have pop the design.
    positions = list(zip(*design_hashes))[0]
    if positions[0] > 1:
        return (design.design_id, 0.0)

    # Count how many consecutive positions there are.
    n_designs: int = 1
    if len(positions) > 1:
        positions = [p if p >= 0 else p + 1 for p in positions]
        for idx in range(1, len(positions)):
            if positions[idx] == positions[idx - 1] + 1:
                n_designs += 1
            else:
                break

    # Cap to the number of designs after which we'll force a reconfiguration.
    max_designs_epoch: int = config["configuration"]["max_designs_epoch"]
    if n_designs > max_designs_epoch:
        n_designs = max_designs_epoch

    return (design.design_id, n_designs / 2 * 900.0)


@check_database
def load_holes(observatory: str):
    """Loads a list holes to ``targetdb.hole``."""

    targetdb.database.become_admin()

    observatory_pk = targetdb.Observatory.get(label=observatory).pk

    row_start = 13
    row_end = -13
    min_cols = 14

    holes = []
    for row in range(row_start, row_end - 1, -1):
        end_col = min_cols + ((row_start - row) if row >= 0 else (row - row_end))
        for col in range(1, end_col + 1, 1):
            if row == 0:
                holeid = f"R0C{col}"
            else:
                holeid = f"R{row:+}C{col}"

            holes.append(
                dict(
                    row=row,
                    column=col,
                    holeid=holeid,
                    observatory_pk=observatory_pk,
                )
            )

    targetdb.Hole.insert(holes).on_conflict(
        conflict_target=[targetdb.Hole.holeid, targetdb.Hole.observatory],
        action="IGNORE",
    ).execute(targetdb.database)


@check_database
def load_fields(
    plan: str,
    files: list[str] | None = None,
    pattern: str | None = None,
    sequential_field: bool = False,
):
    """Loads a series of field.

    Parameters
    ----------
    plan
        The robostrategy run string.
    files
        A list of files to load.
    pattern
        Alternative to ``files``, a pattern to be used with ``glob``
        to retrieve a list of files to load.
    sequential_field
        If `False`, uses the filename to determine the field ID. Otherwise
        sequentially increments the ``field_id`` field starting with the current
        maximum value.
    """

    targetdb.database.become_admin()

    if files is None and pattern:
        files = list(glob(pattern))

    if files is None or len(files) == 0:
        raise ValueError("No files provided.")

    version = targetdb.Version.get_or_create(
        plan=plan,
        target_selection=False,
        robostrategy=True,
    )

    hole_ids = pandas.DataFrame(
        list(
            targetdb.Hole.select(
                targetdb.Hole.pk,
                targetdb.Hole.holeid,
                targetdb.Observatory.label,
            )
            .join(targetdb.Observatory)
            .where(targetdb.Hole.holeid.is_null(False))
            .tuples()
        ),
        columns=["pk", "holeid", "observatory"],
    ).set_index("holeid")

    for file_ in files:
        hdul = fits.open(file_)

        # Create field
        if sequential_field is False:
            field_id = int(file_.split("-")[-1][:-5])
        else:
            field_id = targetdb.Field.select(
                peewee.fn.max(targetdb.Field.field_id)
            ).scalar()
            field_id += 1

        observatory = hdul[0].header["OBS"].upper()
        racen = float(hdul[0].header["RACEN"])
        deccen = float(hdul[0].header["DECCEN"])
        PA = float(hdul[0].header["PA"])
        field_cadence = hdul[0].header["FCADENCE"]
        nexp = hdul[0].header["NEXP"]

        insert = targetdb.Field.insert(
            field_id=field_id,
            racen=racen,
            deccen=deccen,
            position_angle=PA,
            version_pk=version[0].pk,
            cadence_pk=targetdb.Cadence.get(label=field_cadence).pk,
            observatory=targetdb.Observatory.get(label=observatory).pk,
        ).on_conflict(
            conflict_target=[targetdb.Field.field_id],
            preserve=[targetdb.Field.field_id],
            update={
                "observatory_pk": peewee.EXCLUDED.observatory_pk,
                "racen": peewee.EXCLUDED.racen,
                "deccen": peewee.EXCLUDED.deccen,
                "position_angle": peewee.EXCLUDED.position_angle,
                "version_pk": peewee.EXCLUDED.version_pk,
                "cadence_pk": peewee.EXCLUDED.cadence_pk,
            },
        )
        field_pk = insert.execute(targetdb.database)

        # Now create a design for each exposure.
        design_modes = hdul[0].header["DESMODE"].split()

        assign_data = table.hstack(
            (
                table.Table(hdul["ASSIGN"].data),
                table.Table(hdul["TARGET"].data),
            )
        )
        assign_data = assign_data[assign_data["assigned"] > 0]
        insert_data = []

        for n in range(nexp):
            design = targetdb.Design(
                exposure=n + 1,
                field_pk=field_pk,
                design_mode_label=design_modes[n],
            )
            design.save()

            if nexp == 1:
                exp_data = assign_data
                holeIDs = exp_data["holeID"].tolist()
            else:
                exp_data = assign_data[assign_data["holeID"][:, n] != " "]
                holeIDs = exp_data["holeID"][:, n].tolist()

            obs_holes = hole_ids.loc[hole_ids.observatory == observatory]
            holeid_pk_exp = obs_holes.loc[holeIDs].pk.tolist()

            insert_data += [
                {
                    "design_id": design.design_id,
                    "hole_pk": holeid_pk_exp[i],
                    "carton_to_target_pk": exp_data["carton_to_target_pk"][i],
                    "instrument_pk": targetdb.Instrument.get(
                        label=exp_data["fiberType_2"][i]
                    ).pk,
                }
                for i in range(len(exp_data))
            ]

        targetdb.Assignment.insert_many(insert_data).execute(targetdb.database)


def match_assignment_hash(design_id1: int, design_id2: int):
    """Checks if the assignment hashes of two designs match."""

    if design_id1 == design_id2:
        return True

    design1 = targetdb.Design.get_by_id(design_id1)
    design2 = targetdb.Design.get_by_id(design_id2)

    return design1.assignment_hash.hex == design2.assignment_hash.hex



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-15
# @Filename: calibration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import struct

from typing import TYPE_CHECKING, Dict, List, Tuple

import numpy

from jaeger import config, log
from jaeger.commands import Command, CommandID
from jaeger.exceptions import JaegerError
from jaeger.maskbits import PositionerStatus as PS
from jaeger.utils import bytes_to_int, int_to_bytes, motor_steps_to_angle


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = [
    "calibrate_positioners",
    "StartDatumCalibration",
    "StartMotorCalibration",
    "StartCoggingCalibration",
    "SaveInternalCalibration",
    "HallOn",
    "HallOff",
]


MOTOR_STEPS = config["positioner"]["motor_steps"]


async def calibrate_positioners(
    fps: FPS,
    axis: str,
    positioner_ids: int | list[int],
    motors: bool = True,
    datums: bool = True,
    cogging: bool = True,
):
    """Runs the calibration process and saves it to the internal memory.

    Parameters
    ----------
    fps
        The instance of `.FPS` that will receive the trajectory.
    axis
        The axis to calibrate, either `"alpha"`, `"beta"`, or `"both"`.
    positioner_id
        The ID of the positioner(s) to calibrate.
    motors
        Whether to perform the motor calibration.
    datums
        Whether to perform the datums calibration.
    cogging
        Whether to perform the cogging calibration (may take more
        than one hour).

    Raises
    ------
    JaegerError
        If encounters a problem during the process.

    Examples
    --------
    ::

        >>> fps = FPS()
        >>> await fps.initialise()

        # Calibrate positioner 31.
        >>> await calibrate_positioner(fps, 31)

    """

    log.info(f"Calibrating positioner(s) {positioner_ids}.")

    if axis not in ["alpha", "beta", "both"]:
        raise JaegerError(f"Invalid axis {axis!r}.")

    if isinstance(positioner_ids, int):
        positioner_ids = [positioner_ids]

    for positioner_id in positioner_ids:
        if positioner_id not in fps.positioners:
            raise JaegerError(f"Positioner {positioner_id} not found.")

    if fps.pollers.running:
        log.debug("Stopping pollers")
        await fps.pollers.stop()

    if motors:
        log.info("Starting motor calibration.")

        if axis == "alpha":
            motor_command = CommandID.START_MOTOR_CALIBRATION_ALPHA
        elif axis == "beta":
            motor_command = CommandID.START_MOTOR_CALIBRATION_BETA
        else:
            motor_command = CommandID.START_MOTOR_CALIBRATION

        cmd = await fps.send_command(motor_command, positioner_ids=positioner_ids)

        if cmd.status.failed:
            raise JaegerError("Motor calibration failed.")

        await asyncio.sleep(1)

        statuses = [
            PS.DISPLACEMENT_COMPLETED,
            PS.MOTOR_ALPHA_CALIBRATED,
            PS.MOTOR_BETA_CALIBRATED,
        ]
        await _wait_status(fps, positioner_ids, statuses)

    else:
        log.warning("Skipping motor calibration.")

    if datums:
        log.info("Starting datum calibration.")

        if axis == "alpha":
            datums_command = CommandID.START_DATUM_CALIBRATION_ALPHA
        elif axis == "beta":
            datums_command = CommandID.START_DATUM_CALIBRATION_BETA
        else:
            datums_command = CommandID.START_DATUM_CALIBRATION

        cmd = await fps.send_command(datums_command, positioner_ids=positioner_ids)

        if cmd.status.failed:
            raise JaegerError("Datum calibration failed.")

        await asyncio.sleep(1)

        statuses = [
            PS.DISPLACEMENT_COMPLETED,
            PS.DATUM_ALPHA_CALIBRATED,
            PS.DATUM_BETA_CALIBRATED,
        ]
        await _wait_status(fps, positioner_ids, statuses)

    else:
        log.warning("Skipping datum calibration.")

    if cogging:
        log.info("Starting cogging calibration.")

        if axis == "alpha":
            cogging_command = CommandID.START_COGGING_CALIBRATION_ALPHA
        elif axis == "beta":
            cogging_command = CommandID.START_COGGING_CALIBRATION_BETA
        else:
            cogging_command = CommandID.START_COGGING_CALIBRATION

        cmd = await fps.send_command(cogging_command, positioner_ids=positioner_ids)

        if cmd.status.failed:
            raise JaegerError("Cogging calibration failed.")

        await asyncio.sleep(1)

        statuses = [PS.COGGING_ALPHA_CALIBRATED, PS.COGGING_BETA_CALIBRATED]
        await _wait_status(fps, positioner_ids, statuses)

    else:
        log.warning("Skipping cogging calibration.")

    if motors or datums or cogging:
        log.info("Saving calibration.")
        cmd = await fps.send_command(
            CommandID.SAVE_INTERNAL_CALIBRATION,
            positioner_ids=positioner_ids,
        )
        if cmd.status.failed:
            raise JaegerError("Saving calibration failed.")

        log.info(f"Positioners {positioner_ids} have been calibrated.")

    return


async def _wait_status(fps: FPS, positioner_ids: list[int], statuses: list[PS]):
    """Waits for status."""

    wait_for_status_tasks = []
    for positioner_id in positioner_ids:
        positioner = fps.positioners[positioner_id]
        wait_for_status_tasks.append(positioner.wait_for_status(statuses))

    await asyncio.gather(*wait_for_status_tasks)


class StartDatumCalibration(Command):
    """Indicates that the transmission for the trajectory has ended."""

    command_id = CommandID.START_DATUM_CALIBRATION
    broadcastable = False
    move_command = True


class StartMotorCalibration(Command):
    """Aborts sending a trajectory."""

    command_id = CommandID.START_MOTOR_CALIBRATION
    broadcastable = False
    move_command = True


class StartCoggingCalibration(Command):
    """Starts the trajectories."""

    command_id = CommandID.START_COGGING_CALIBRATION
    broadcastable = False
    move_command = True


class SaveInternalCalibration(Command):
    """Stop the trajectories."""

    command_id = CommandID.SAVE_INTERNAL_CALIBRATION
    broadcastable = False
    move_command = False


class GetOffset(Command):
    """Gets the motor offsets."""

    command_id = CommandID.GET_OFFSETS
    broadcastable = False
    safe = True

    def get_replies(self) -> Dict[int, numpy.ndarray]:
        return self.get_offsets()

    def get_offsets(self) -> Dict[int, numpy.ndarray]:
        """Returns the alpha and beta offsets, in degrees.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.

        """

        offsets = {}
        for reply in self.replies:
            pid = reply.positioner_id
            data = reply.data

            alpha = bytes_to_int(data[0:4], dtype="i4")
            beta = bytes_to_int(data[4:], dtype="i4")

            offsets[pid] = numpy.array(motor_steps_to_angle(alpha, beta))

        return offsets


class SetOffsets(Command):
    """Sets the motor offsets."""

    command_id = CommandID.SET_OFFSETS
    broadcastable = False
    safe = True
    move_command = False

    def __init__(
        self,
        positioner_ids: int | List[int],
        alpha=None,
        beta=None,
        **kwargs,
    ):

        if alpha is not None and beta is not None:
            alpha_steps, beta_steps = motor_steps_to_angle(alpha, beta, inverse=True)

            data = int_to_bytes(int(alpha_steps)) + int_to_bytes(int(beta_steps))
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


class HallOn(Command):
    """Turns hall sensors ON."""

    command_id = CommandID.HALL_ON
    broadcastable = False
    move_command = False
    safe = True


class HallOff(Command):
    """Turns hall sensors ON."""

    command_id = CommandID.HALL_OFF
    broadcastable = False
    move_command = False
    safe = True


class SetHoldingCurrents(Command):
    """Sets the motors holding currents."""

    command_id = CommandID.SET_HOLDING_CURRENT
    broadcastable = False
    safe = True
    move_command = False

    def __init__(self, positioner_ids, alpha=None, beta=None, **kwargs):

        if alpha is not None and beta is not None:
            data = int_to_bytes(int(alpha)) + int_to_bytes(int(beta))
            kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


class GetHoldingCurrents(Command):
    """Gets the motor offsets."""

    command_id = CommandID.GET_HOLDING_CURRENT
    broadcastable = False
    safe = True

    def get_replies(self) -> Dict[int, numpy.ndarray]:
        return self.get_holding_currents()

    def get_holding_currents(self) -> Dict[int, numpy.ndarray]:
        """Returns the alpha and beta holding currents, in percent.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.

        """

        currents = {}
        for reply in self.replies:
            data = reply.data

            alpha = bytes_to_int(data[0:4], dtype="i4")
            beta = bytes_to_int(data[4:], dtype="i4")

            currents[reply.positioner_id] = numpy.array([alpha, beta])

        return currents


class PreciseMoveAlphaOn(Command):
    """Turns precise move on alpha ON."""

    command_id = CommandID.SWITCH_ON_PRECISE_MOVE_ALPHA
    broadcastable = False
    move_command = False
    safe = True


class PreciseMoveAlphaOff(Command):
    """Turns precise move on alpha OFF."""

    command_id = CommandID.SWITCH_OFF_PRECISE_MOVE_ALPHA
    broadcastable = False
    move_command = False
    safe = True


class PreciseMoveBetaOn(Command):
    """Turns precise move on beta ON."""

    command_id = CommandID.SWITCH_ON_PRECISE_MOVE_BETA
    broadcastable = False
    move_command = False
    safe = True


class PreciseMoveBetaOff(Command):
    """Turns precise move on beta OFF."""

    command_id = CommandID.SWITCH_OFF_PRECISE_MOVE_BETA
    broadcastable = False
    move_command = False
    safe = True


class SetIncreaseCollisionMargin(Command):
    """Sets the buffer for collision margin."""

    command_id = CommandID.SET_INCREASE_COLLISION_MARGIN
    broadcastable = False
    move_command = False
    safe = False

    def __init__(self, positioner_ids, margin: int, **kwargs):

        data = int_to_bytes(int(margin), dtype="i4")
        kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)


class GetAlphaHallCalibration(Command):

    command_id = CommandID.GET_ALPHA_HALL_CALIB
    broadcastable = False
    move_command = False
    safe = True

    def get_replies(self) -> Dict[int, Tuple[int, int, int, int]]:
        return self.get_values()

    def get_values(self) -> dict[int, Tuple[int, int, int, int]]:
        """Returns the ``maxA, maxB, minA, minB`` values."""

        values = {}

        for reply in self.replies:
            values[reply.positioner_id] = struct.unpack("HHHH", reply.data)

        return values


class GetBetaHallCalibration(Command):

    command_id = CommandID.GET_BETA_HALL_CALIB
    broadcastable = False
    move_command = False
    safe = True

    def get_replies(self) -> Dict[int, Tuple[int, int, int, int]]:
        return self.get_values()

    def get_values(self) -> dict[int, Tuple[int, int, int, int]]:
        """Returns the ``maxA, maxB, minA, minB`` values."""

        values = {}

        for reply in self.replies:
            values[reply.positioner_id] = struct.unpack("HHHH", reply.data)

        return values


class GetHallCalibrationError(Command):

    command_id = CommandID.GET_HALL_CALIB_ERROR
    broadcastable = False
    move_command = False
    safe = True

    def get_replies(self) -> Dict[int, Tuple[int, int]]:
        return self.get_values()

    def get_values(self) -> dict[int, Tuple[int, int]]:
        """Returns the alpha and beta error values."""

        values = {}

        for reply in self.replies:
            values[reply.positioner_id] = struct.unpack("ii", reply.data)

        return values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-31
# @Filename: disable.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from jaeger import config

from . import JaegerCommandType, jaeger_parser


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = ["disable", "enable"]


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int, required=False)
async def disable(
    command: JaegerCommandType,
    fps: FPS,
    positioner_id: int | None = None,
):
    """Disables a positioner"""

    permanently_disabled = config["fps"]["disabled_positioners"]
    if config["fps"]["offline_positioners"] is not None:
        permanently_disabled += list(config["fps"]["offline_positioners"].keys())

    if positioner_id:

        if positioner_id not in fps:
            return command.fail(f"Positioner {positioner_id} is not in the array.")

        positioner = fps.positioners[positioner_id]

        if positioner.disabled:
            command.warning(f"Positioner {positioner_id} is already disabled.")

        positioner.disabled = True
        fps.disabled.add(positioner.positioner_id)

    manually_disabled: list[int] = []
    for positioner in fps.positioners.values():
        if positioner.offline or positioner.disabled:
            if positioner.positioner_id not in permanently_disabled:
                manually_disabled.append(positioner.positioner_id)

    return command.finish(
        permanently_disabled=list(sorted(permanently_disabled)),
        manually_disabled=list(sorted(manually_disabled)),
    )


@jaeger_parser.command()
@click.argument("POSITIONER-ID", type=int)
async def enable(command: JaegerCommandType, fps: FPS, positioner_id: int):
    """Enables a positioner"""

    permanently_disabled = config["fps"]["disabled_positioners"]
    if config["fps"]["offline_positioners"] is not None:
        permanently_disabled += list(config["fps"]["offline_positioners"].keys())

    if positioner_id not in fps:
        return command.fail(f"Positioner {positioner_id} is not in the array.")

    if positioner_id in permanently_disabled:
        return command.fail(
            f"Positioner {positioner_id} is permanently "
            "disabled and cannot be enabled."
        )

    positioner = fps.positioners[positioner_id]

    if positioner.disabled is False and positioner.offline is False:
        command.warning(f"Positioner {positioner_id} is not disabled.")

    positioner.disabled = False
    positioner.offline = False

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-07-11
# @Filename: virtual.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import Dict, List

from jaeger.interfaces.bus import BusABC

from .message import Message


queues: Dict[str, List[asyncio.Queue]] = {}


class VirtualBus(BusABC):
    """A class implementing a virtual CAN bus that listens to messages on a channel."""

    def __init__(self, channel: str):

        self.channel = channel

        self.queue: asyncio.Queue[Message] = asyncio.Queue()

        if self.channel not in queues:
            queues[self.channel] = [self.queue]
        else:
            queues[self.channel].append(self.queue)

    def send(self, msg: Message):
        """Send message to the virtual bus (self does not receive a copy)."""

        for queue in queues[self.channel]:
            if queue is self.queue:
                continue
            queue.put_nowait(msg)

    async def get(self):
        """Get messages from the bus."""

        msg = await self.queue.get()

        return msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-24
# @Filename: snapshot.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import jaeger_parser


if TYPE_CHECKING:
    from clu.command import Command

    from jaeger import FPS
    from jaeger.actor import JaegerActor


__all__ = ["snapshot"]


@jaeger_parser.command()
@click.argument("PATH", required=False, type=click.Path(exists=False, dir_okay=False))
async def snapshot(command: Command[JaegerActor], fps: FPS, path: str | None = None):
    """Takes a snapshot image."""

    if path is not None:
        path = str(path)

    try:
        filename = await fps.save_snapshot(path, write_to_actor=False)
    except Exception as err:
        return command.fail(f"Snapshot failed with error: {err}")

    return command.finish(snapshot=filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-05-04
# @Filename: plotting.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

import matplotlib.pyplot as plt
import numpy
import pandas
import seaborn

from jaeger.target.configuration import BaseConfiguration


__all__ = ["plot_fvc_distances"]


def _plot_wok_distance(data_F: pandas.DataFrame, ax: plt.Axes, is_dither: bool = False):

    colours = ["g", "r", "b"]
    for ii, fibre in enumerate(["Metrology", "APOGEE", "BOSS"]):

        data_fibre = data_F.loc[pandas.IndexSlice[:, fibre], :].copy()

        if fibre != "Metrology" and is_dither is False:
            data_fibre = data_fibre.loc[
                (data_fibre.assigned == 1) & (data_fibre.on_target == 1), :
            ]
        if len(data_fibre) == 0:
            continue

        wok_distance = data_fibre.wok_distance * 1000.0
        wok_distance_bin = numpy.histogram(wok_distance, bins=numpy.arange(0, 105, 5))

        perc_90 = numpy.percentile(wok_distance, 90)
        ax.plot(
            wok_distance_bin[1][1:] - wok_distance_bin[1][1] / 2.0,  # type: ignore
            wok_distance_bin[0],
            linestyle="-",
            color=colours[ii],
            label=rf"{fibre}: {perc_90:.1f} $\mu$m",
            zorder=10,
        )

        if fibre == "Metrology":
            for perc_q in [90]:
                perc = numpy.percentile(wok_distance, perc_q)
                ax.axvline(x=perc, color="k", linestyle="--", linewidth=0.5, zorder=0)

        ax.set_xlim(0, 100)

        ax.set_xlabel("Wok distance [microns]")
        ax.set_ylabel("Number")

        ax.set_title("Wok distance")

        ax.legend()


def _plot_sky_distance(
    data_F: pandas.DataFrame,
    ax: plt.Axes,
    column: str,
    is_dither: bool = False,
    plot_metrology: bool = True,
    title: str = "Sky distance",
):

    colours = ["g", "r", "b"]
    for ii, fibre in enumerate(["Metrology", "APOGEE", "BOSS"]):
        if fibre == "Metrology" and not plot_metrology:
            continue

        data_fibre = data_F.loc[pandas.IndexSlice[:, fibre], :].copy()

        if fibre != "Metrology" and is_dither is False:
            data_fibre = data_fibre.loc[
                (data_fibre.assigned == 1) & (data_fibre.on_target == 1), :
            ]
        if len(data_fibre) == 0:
            continue

        sky_distance = data_fibre[column]
        sky_distance_bin = numpy.histogram(sky_distance, bins=numpy.arange(0, 1.5, 0.1))

        perc_90 = numpy.percentile(sky_distance, 90)
        ax.plot(
            sky_distance_bin[1][1:] - sky_distance_bin[1][1] / 2.0,  # type: ignore
            sky_distance_bin[0],
            linestyle="-",
            color=colours[ii],
            label=f"{fibre}: {perc_90:.2f} arcsec",
            zorder=10,
        )

        if fibre == "Metrology":
            for perc_q in [90]:
                perc = numpy.percentile(sky_distance, perc_q)
                ax.axvline(x=perc, color="k", linestyle="--", linewidth=0.5, zorder=0)

        ax.set_xlim(0, 1.5)

        ax.set_xlabel("Sky distance [arcsec]")
        ax.set_ylabel("Number")

        ax.set_title(title)

        ax.legend()


def _plot_sky_quiver(data_F: pandas.DataFrame, ax: plt.Axes, is_dither: bool = False):

    colours = ["r", "b"]
    for ii, fibre in enumerate(["APOGEE", "BOSS"]):

        data_fibre = data_F.loc[pandas.IndexSlice[:, fibre], :].copy()

        if is_dither is False:
            data_fibre = data_fibre.loc[
                (data_fibre.assigned == 1) & (data_fibre.on_target == 1), :
            ]
        if len(data_fibre) == 0:
            continue

        q = ax.quiver(
            data_fibre.ra_epoch,
            data_fibre.dec_epoch,
            data_fibre.ra_distance,
            data_fibre.dec_distance,
            color=colours[ii],
            label=fibre,
        )

        ax.quiverkey(
            q,
            X=0.05,
            Y=0.05 * (ii + 1),
            U=0.2,
            label="0.2 arcsec",
            labelpos="E",
        )

        ax.set_xlabel("Right Ascension")
        ax.set_ylabel("Declination")

        ax.set_title("confSummary(ra/dec) - confSummaryF(ra/dec)")

        ax.legend()


def plot_fvc_distances(
    configuration: BaseConfiguration,
    fibre_data_F: pandas.DataFrame,
    path: str | pathlib.Path | None = None,
):
    """Plots several panels with FVC analysis and statistics.

    Parameters
    ----------
    fibre_data
        Data frame with the targetting data.
    fibre_data_F
        Data frame with the measured values from the FVC.

    Returns
    -------
    figure
        Matplotlib figure with the plots.

    """

    data = configuration.assignment_data.fibre_table.copy()
    data = data.reset_index().set_index(["positioner_id", "fibre_type"])

    data_F = fibre_data_F.copy()
    data_F = data_F.reset_index().set_index(["positioner_id", "fibre_type"])

    is_dither = configuration.is_dither

    data = data.loc[data_F.index, :]

    data_F["xwok_distance"] = data.xwok - data_F.xwok
    data_F["ywok_distance"] = data.ywok - data_F.ywok
    data_F["wok_distance"] = numpy.hypot(data_F.xwok_distance, data_F.ywok_distance)

    deccen = configuration.assignment_data.boresight[0][1]
    cos_dec = numpy.cos(numpy.deg2rad(float(deccen)))

    data_F["ra_distance"] = (data.ra_epoch - data_F.ra_epoch) * cos_dec * 3600.0
    data_F["dec_distance"] = (data.dec_epoch - data_F.dec_epoch) * 3600.0
    data_F["sky_distance"] = numpy.hypot(data_F.ra_distance, data_F.dec_distance)

    if not is_dither:
        data_F["racat_distance"] = (data_F.ra_icrs - data_F.ra_epoch) * cos_dec * 3600.0
        data_F["deccat_distance"] = (data_F.dec_icrs - data_F.dec_epoch) * 3600.0
        data_F["skycat_distance"] = numpy.hypot(
            data_F.racat_distance, data_F.deccat_distance
        )

    with plt.ioff():  # type: ignore

        seaborn.set_theme()

        fig, axes = plt.subplots(1, 3, figsize=(30, 10))

        if not is_dither:
            data_F = data_F.groupby("positioner_id").filter(
                lambda g: g.assigned.any() & g.on_target.any() & g.valid.all()
            )

        assert isinstance(axes, numpy.ndarray)

        _plot_wok_distance(data_F, axes[0])

        _plot_sky_distance(
            data_F,
            axes[1],
            "sky_distance",
            is_dither=is_dither,
            plot_metrology=True,
            title="Sky distance (ra/dec vs ra/dec)",
        )

        # if not is_dither:
        #     _plot_sky_distance(
        #         data_F,
        #         axes[1, 0],
        #         "skycat_distance",
        #         is_dither=is_dither,
        #         plot_metrology=False,
        #         title="Sky distance (ra/dec vs racat/deccat)",
        #     )

        _plot_sky_quiver(data_F, axes[2], is_dither=is_dither)

        fig.suptitle(
            f"Configuration ID: {configuration.configuration_id}"
            + (" (dithered)" if is_dither else "")
        )

        plt.tight_layout()

        if path:
            fig.savefig(str(path))
            plt.close(fig)

        seaborn.reset_defaults()

    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-08-27
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import binascii
import collections
import logging
import sys
import time
import warnings

from typing import Any, Callable, Dict, List, Optional, Union

from jaeger import can_log, config, log, maskbits
from jaeger.exceptions import CommandError, JaegerError, JaegerUserWarning
from jaeger.interfaces import BusABC, Message
from jaeger.maskbits import CommandStatus, ResponseCode
from jaeger.utils import StatusMixIn, get_identifier, parse_identifier

from . import CommandID


__all__ = ["SuperMessage", "Command", "EmptyPool"]


# A pool of UIDs that can be assigned to each command for a given command_id.
# The format of the pool is UID_POOL[command_id][positioner_id] so that each
# positioner has uid_bits (64) messages for each command id. UID=0 is always
# reserved for broadcasts.
UID_POOL = collections.defaultdict(dict)

# Starting value for command UID.
COMMAND_UID = 0


class SuperMessage(Message):
    """An extended CAN ``Message`` class.

    Expands the ``Message`` class to handle custom arbitration IDs for
    extended frames.

    Parameters
    ----------
    command
        The command associated with this message.
    data
        Payload to pass to ``Message``.
    positioner_id
        The positioner to which the message will be sent (0 for broadcast).
    uid
        The unique identifier for this message.
    extended_id
        Whether the id is an 11 bit (False) or 29 bit (True) address.

    """

    def __init__(
        self,
        command: Command,
        data: bytearray = bytearray([]),
        positioner_id: int = 0,
        uid: int = 0,
        response_code: int = 0,
        extended_id: bool = True,
    ):

        self.command = command
        self.positioner_id = positioner_id
        self.uid = uid

        uid_bits = config["positioner"]["uid_bits"]
        max_uid = 2**uid_bits
        assert self.uid < max_uid, f"UID must be <= {max_uid}."

        if extended_id:
            arbitration_id = get_identifier(
                positioner_id,
                int(command.command_id),
                uid=self.uid,
                response_code=response_code,
            )
        else:
            arbitration_id = positioner_id

        Message.__init__(
            self,
            data=data,
            arbitration_id=arbitration_id,
            is_extended_id=extended_id,
        )


class Reply(object):
    """Parses a reply message.

    Parameters
    ----------
    message
        The received message
    command
        The `.Command` to which this message is replying.

    """

    def __init__(self, message: Message, command: Optional[Command] = None):

        assert isinstance(message, Message), "invalid message"

        #: The command for which this reply is intended.
        self.command = command

        #: The raw ``Message``.
        self.message = message

        #: The data from the message.
        self.data = message.data

        #: The `~.maskbits.ResponseCode` bit returned by the reply.
        self.response_code: maskbits.ResponseCode

        #: The UID of the message this reply is for.
        self.uid: int

        (
            self.positioner_id,
            reply_cmd_id,
            self.uid,
            self.response_code,
        ) = parse_identifier(message.arbitration_id)

        if command is not None:
            assert command.command_id == reply_cmd_id, (
                f"Command command_id={command.command_id} and "
                f"reply command_id={reply_cmd_id} do not match"
            )

        self.command_id = CommandID(reply_cmd_id)

    def __repr__(self):
        command_name = self.command.command_id.name if self.command else "NONE"
        return (
            f"<Reply (command_id={command_name!r}, "
            f"positioner_id={self.positioner_id}, "
            f"response_code={self.response_code.name!r})>"
        )


class EmptyPool(CommandError):
    pass


data_co = Union[None, bytearray, List[bytearray]]


if sys.version_info >= (3, 9):
    Future_co = asyncio.Future["Command"]
else:
    Future_co = asyncio.Future


class Command(StatusMixIn[CommandStatus], Future_co):
    """A command to be sent to the CAN controller.

    Implements a base class to define CAN commands to interact with the
    positioner. Commands can be composed of single or multiple messages.
    When sending a command to the bus, the first message is written to,
    then asynchronously waits for a confirmation that the message has been
    received before sending the following message. If any of the messages
    returns an error code the command is failed.

    `.Command` subclasses from `.StatusMixIn` and `.status_callback` gets
    called when the status changes.

    `.Command` is a `~asyncio.Future` and must be awaited. The
    `~asyncio.Future` is done when `~.Command.finish_command` is called,
    which happens when the status is marked done or cancelled or when the
    command timeouts.

    Commands sent to a single positioner are marked done when a reply is
    received from the same positioner for the given command, or when it
    times out. Broadcast commands only get marked done by timing out or
    manually.

    Parameters
    ----------
    positioner_ids
        The id or list of ids of the robot(s) to which this command will be
        sent. Use ``positioner_ids=0`` to broadcast to all robots.
    loop
        The running event loop, or uses `~asyncio.get_event_loop`.
    timeout
        Time after which the command will be marked done when not all the
        positioners have replies. If `None`, the default timeout will be used.
        If timeout is a negative number, the command won't timeout until all
        the positioners have replied.
    done_callback
        A function to call when the command has been successfully completed.
    n_positioners
        If the command is a broadcast, the number of positioners that should
        reply. If defined, the command will be done once as many positioners
        have replied. Ignored for non-broadcasts.
    data
        The data to pass to the messages. If a list, each element will be
        sent to each positioner as a message. It can also be a dictionary of
        lists in which the key is the positioner to which to send the data.
    ignore_unknown
        Ignores ``UNKNOWN_COMMAND`` replies from positioners that do now
        support this command.

    """

    #: The id of the command.
    command_id: CommandID
    #: Whether the command can be broadcast to all robots.
    broadcastable: bool = False
    #: The default timeout for this command.
    timeout: float = 5
    #: Whether it's safe to execute this command when the FPS is locked.
    safe = False
    #: Whether this command produces a positioner move.
    move_command = False
    #: Whether the command is safe to be issues in bootloader mode.
    bootloader = False

    def __init__(
        self,
        positioner_ids: int | List[int],
        timeout: Optional[float] = None,
        done_callback: Optional[Callable] = None,
        n_positioners: Optional[int] = None,
        data: Union[None, data_co, Dict[int, data_co]] = None,
        ignore_unknown: bool = True,
    ):

        global COMMAND_UID

        assert self.broadcastable is not None, "broadcastable not set"
        assert self.command_id is not None, "command_id not set"

        if isinstance(positioner_ids, (list, tuple)):
            self.positioner_ids = list(positioner_ids)
            if len(positioner_ids) != len(set(positioner_ids)):
                raise JaegerError("The list of positioner_ids must be unique.")
            if len(positioner_ids) > 1 and 0 in positioner_ids:
                raise JaegerError("Broadcasts cannot be mixed with other positioners.")
        else:
            self.positioner_ids = [positioner_ids]

        if self.is_broadcast and self.broadcastable is False:
            raise JaegerError(f"Command {self.command_id.name} cannot be broadcast.")

        #: The data payload for the messages to send.
        if data is None:
            self.data = {pid: [bytearray()] for pid in self.positioner_ids}
        elif isinstance(data, bytearray):
            self.data = {pid: [data] for pid in self.positioner_ids}
        elif isinstance(data, (list, tuple)):
            self.data = {pid: data for pid in self.positioner_ids}
        elif isinstance(data, dict):
            self.data = {}
            for pid, value in data.items():
                if value is None:
                    self.data[pid] = [bytearray()]
                elif isinstance(value, (list, tuple)):
                    self.data[pid] = value
                elif isinstance(value, bytearray):
                    self.data[pid] = [value]
                else:
                    raise ValueError(f"Invalid data {value!r}.")
        else:
            raise ValueError(f"Invalid data {data!r}.")

        if self.is_broadcast:
            if len(self.data[0]) > 1:
                raise CommandError("Broadcasts can only include a single data packet.")

        # Number of replies expected unless broadcast.
        if self.is_broadcast:
            if n_positioners:
                self._n_replies = len(self.data[0]) * n_positioners
            else:
                self._n_replies = None
        else:
            self._n_replies = sum([len(value) for value in self.data.values()])

        if timeout is None:
            pass
        else:
            self.timeout = timeout
            if self.timeout < 0 and self._n_replies is None:
                raise CommandError(
                    "In a broadcast a timeout is required unless n_positioners is set."
                )

        #: A list of messages with the responses to this command.
        self.replies: List[Reply] = []

        # Messages sent.
        self.messages = []
        self.message_uids = []

        # Generate a UUID for this command.
        self.command_uid = COMMAND_UID
        COMMAND_UID += 1

        # Starting and end time
        self.start_time: float | None = None
        self.end_time: float | None = None

        # If this is the first time we run this command the pool will be empty.
        uid_bits = config["positioner"]["uid_bits"]
        command_pool = UID_POOL[self.command_id]
        if not self.is_broadcast:
            for pid in self.positioner_ids:
                if pid not in command_pool:
                    command_pool[pid] = set(range(1, 2**uid_bits))
        else:
            if 0 not in command_pool:
                command_pool[0] = set([0])

        # What interface and bus this command should be sent to. Only relevant
        # for multibus interfaces. To be filled by the FPS class when queueing
        # the command.
        self._interfaces: Optional[List[BusABC]] = []
        self._bus: Optional[int] = None

        self._done_callback = done_callback

        self._timeout_handle = None

        self._ignore_unknown = ignore_unknown
        self.loop = asyncio.get_event_loop()

        StatusMixIn.__init__(
            self,
            maskbit_flags=CommandStatus,
            initial_status=CommandStatus.READY,
            callback_func=self.status_callback,
        )

        asyncio.Future.__init__(self)

    def __repr__(self):
        return (
            f"<Command {self.command_id.name} "
            f"(positioner_ids={self.positioner_ids!r}, "
            f"status={self.status.name!r})>"
        )

    def _log(
        self,
        msg,
        level=logging.DEBUG,
        command_id=None,
        positioner_ids=None,
        logs=[can_log],
    ):
        """Logs a message."""

        command_id = command_id or self.command_id
        c_name = command_id.name
        pid = positioner_ids or self.positioner_ids

        msg = f"[{c_name}, {pid}, {self.command_uid!s}]: " + msg

        for ll in logs:
            ll.log(level, msg)

    @property
    def is_broadcast(self):
        """Returns `True` if the command is a broadcast."""

        return self.positioner_ids == [0]

    @property
    def name(self):
        """Returns the name of this command."""

        return CommandID(self.command_id).name

    def _check_replies(self):
        """Checks if the UIDs of the replies match the messages."""

        sent_uids = self.message_uids
        replies_uids = [reply.uid for reply in self.replies]

        if self.is_broadcast:
            if self._n_replies is None:  # This means it will timeout.
                return None
            else:
                sent_uids = [0] * self._n_replies

        assert self._n_replies, "_n_replies must be set."

        if len(self.replies) < self._n_replies:
            return None

        if len(self.replies) > self._n_replies:
            self._log(
                "command received more replies than messages. "
                "This should not be possible.",
                level=logging.ERROR,
            )
            return False

        # Compares each message-reply UID.
        if sorted(sent_uids) != sorted(replies_uids):
            self._log(
                "the UIDs of the messages and replies do not match.",
                level=logging.ERROR,
            )
            return False

        return True

    async def process_reply(self, reply_message):
        """Watches the reply queue."""

        reply = Reply(reply_message, command=self)

        # Return the UID to the pool.
        if not self.is_broadcast:
            UID_POOL[self.command_id][reply.positioner_id].add(reply.uid)

        pid = reply.positioner_id

        if self.status == CommandStatus.TIMEDOUT:
            self._log(
                f"received a reply from {pid} but the command has already timed out.",
                level=logging.ERROR,
                logs=[log, can_log],
            )
            return
        elif self.status == CommandStatus.CANCELLED:
            return
        elif self.status != CommandStatus.RUNNING:
            self._log(
                f"received a reply from {pid} but command is not running",
                level=logging.ERROR,
                logs=[log, can_log],
            )
            return

        if not self.is_broadcast:
            if reply.positioner_id not in self.positioner_ids:
                self._log(
                    f"received a reply from {pid} from a non-commanded positioner.",
                    level=logging.ERROR,
                    logs=[log, can_log],
                )
                return

        self.replies.append(reply)

        data_hex = binascii.hexlify(reply.data).decode()
        self._log(
            f"positioner {reply.positioner_id} replied with "
            f"id={reply.message.arbitration_id}, "
            f"UID={reply.uid}, "
            f"code={reply.response_code.name!r}, "
            f"data={data_hex!r}"
        )

        code = reply.response_code
        COMMAND_ACCEPTED = ResponseCode.COMMAND_ACCEPTED
        UNKNOWN_COMMAND = ResponseCode.UNKNOWN_COMMAND
        if code != COMMAND_ACCEPTED:
            if not self._ignore_unknown or code != UNKNOWN_COMMAND:
                warnings.warn(
                    f"Positioner {reply.positioner_id} replied to {self.name} "
                    f"UID={self.command_uid} with {code.name!r}.",
                    JaegerUserWarning,
                )

        reply_status = self._check_replies()

        # If reply_status is True then a reply from each commanded positioner has
        # been received. If they are all COMMAND_ACCEPTED, mark the command as done.
        # If some are not COMMAND_ACCEPTED, check the invalid replies are all
        # UNKNOWN_COMMAND. In that case, if we are ignoring those, still mark as done.
        # Otherwise finish as failed.
        # If reply_status is False, that means we have received replies with UIDs that
        # do not match the UID of this command. This should not happens and it's most
        # likely a bug in the code.
        # If reply_status is None, we haven't yet matched the expected number of
        # replies and we just return.

        if reply_status is True:
            reply_codes = [reply.response_code for reply in self.replies]
            invalid = [code for code in reply_codes if code != COMMAND_ACCEPTED]
            ignore_unknown = self._ignore_unknown
            if not all([code == COMMAND_ACCEPTED for code in reply_codes]):
                if all([inv == UNKNOWN_COMMAND for inv in invalid]) and ignore_unknown:
                    self.finish_command(CommandStatus.DONE)
                else:
                    self.finish_command(CommandStatus.FAILED)
            else:
                self.finish_command(CommandStatus.DONE)
        elif reply_status is False:
            self.finish_command(CommandStatus.FAILED)
        else:
            return

    def finish_command(self, status: CommandStatus, silent: bool = False):
        """Finishes a command, marking the Future as done.

        Parameters
        ----------
        status
            The status to set the command to. If `None` the command will be set
            to `~.CommandStatus.DONE` if one reply for each message has been
            received, `~.CommandStatus.FAILED` otherwise.
        silent
            If `True`, issues error log messages as debug.

        """

        if self._timeout_handle:
            self._timeout_handle.cancel()

        self._status = status

        if not self.done():
            level = logging.WARNING if not silent else logging.DEBUG
            if not self.is_broadcast and self.status == CommandStatus.TIMEDOUT:
                self._log("this command timed out and it is not a broadcast.", level)
            elif self.status == CommandStatus.CANCELLED:
                self._log("command has been cancelled.", logging.DEBUG)
            elif self.status.failed:
                level = logging.ERROR if not silent else logging.DEBUG
                self._log(f"command finished with status {self.status.name!r}", level)

            # For good measure we return all the UIDs
            if self.is_broadcast:
                UID_POOL[self.command_id][0].add(0)
            else:
                for message in self.messages:
                    UID_POOL[self.command_id][message.positioner_id].add(message.uid)

            self.set_result(self)
            self.end_time = time.time()

            is_done = self.status in [CommandStatus.TIMEDOUT, CommandStatus.DONE]

            if is_done and self._done_callback:
                if asyncio.iscoroutinefunction(self._done_callback):
                    asyncio.create_task(self._done_callback())
                else:
                    self._done_callback()

            self._log(f"finished command with status {self.status.name!r}")

    def status_callback(self):
        """Callback for change status.

        When the status gets set to `.CommandStatus.RUNNING` starts a timer
        that marks the command as done after `.timeout`.

        """

        self._log(f"status changed to {self.status.name}")

        if self.status == CommandStatus.RUNNING:
            self.start_time = time.time()
            if self.timeout < 0:
                pass
            elif self.timeout == 0:
                self.finish_command(CommandStatus.TIMEDOUT)
            else:
                self._timeout_handle = self.loop.call_later(
                    self.timeout,
                    self.finish_command,
                    CommandStatus.TIMEDOUT,
                )
        elif self.status.is_done and not self.done():
            self.finish_command(self.status)

    def _generate_messages_internal(self, data: Optional[List[bytearray]] = None):
        """Generates the list of messages to send to the bus for this command.

        This method is called by `.get_messages` and can be overridden in
        subclasses. Do not override `.get_messages` directly.

        """

        cid = self.command_id

        messages: List[SuperMessage] = []

        for pid in self.data:

            pid_data = self.data[pid]

            for d in pid_data:

                try:
                    uid = UID_POOL[cid][pid].pop()
                except KeyError:
                    # Before failing, put back the UIDs of the other messages
                    for message in messages:
                        UID_POOL[cid][pid].add(message.uid)
                    raise EmptyPool("no UIDs left in the pool.")

                messages.append(SuperMessage(self, positioner_id=pid, uid=uid, data=d))

        return messages

    def get_messages(self, data=None):
        """Returns the list of messages associated with this command.

        Unless overridden, returns a single message with the associated data.

        """

        if len(self.messages) > 0:
            raise CommandError("Messages have already been sent.")

        messages = self._generate_messages_internal(data=data)

        self.messages = messages
        self.message_uids = [message.uid for message in messages]

        return messages

    def get_replies(self) -> Dict[int, Any]:
        """Returns the formatted replies as a dictionary.

        The values returned will depend on the specific command.

        """

        return {}

    def cancel(self, silent=False, msg=None):
        """Cancels a command, stopping the reply queue watcher."""

        self.finish_command(CommandStatus.CANCELLED, silent=silent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-09-06
# @Filename: fps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import pathlib
import warnings
from dataclasses import dataclass
from glob import glob

from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy
from astropy.time import Time
from zc.lockfile import LockFile

import jaeger
from jaeger import can_log, config, log, start_file_loggers
from jaeger.can import JaegerCAN
from jaeger.commands import (
    Command,
    CommandID,
    GetFirmwareVersion,
    goto,
    send_trajectory,
)
from jaeger.exceptions import (
    FPSLockedError,
    JaegerError,
    JaegerUserWarning,
    PositionerError,
)
from jaeger.ieb import IEB
from jaeger.interfaces import BusABC
from jaeger.maskbits import FPSStatus, PositionerStatus
from jaeger.positioner import Positioner
from jaeger.utils import Poller, PollerList


if TYPE_CHECKING:
    from matplotlib.axes import Axes

    from jaeger.target.configuration import BaseConfiguration


try:
    from coordio import calibration
except ImportError:
    calibration = None


__all__ = ["BaseFPS", "FPS"]


MIN_BETA = 160
LOCK_FILE = "/var/tmp/sdss/jaeger.lock"

FPS_CO = Union["BaseFPS", "FPS"]
FPS_T = TypeVar("FPS_T", bound="BaseFPS")


class BaseFPS(Dict[int, Positioner], Generic[FPS_T]):
    """A class describing the Focal Plane System.

    This class includes methods to read the layout and construct positioner
    objects and can be used by the real `FPS` class or the
    `~jaeger.testing.VirtualFPS`.

    `.BaseFPS` instances are singletons in the sense that one cannot instantiate
    more than one. An error is raise if ``__new__`` is called with an existing
    instance. To retrieve the running instance, use `.get_instance`.

    Attributes
    ----------
    positioner_class
        The class to be used to create a new positioner. In principle this will
        be `.Positioner` but it may be different if the positioners are created
        for a `~jaeger.testing.VirtualFPS`.

    """

    positioner_class: ClassVar[Type[Positioner]] = Positioner
    _instance: dict[Type[BaseFPS], FPS_T] = {}

    initialised: bool

    def __new__(cls: Type[FPS_T], *args, **kwargs):

        if cls in cls._instance and kwargs == {}:
            raise JaegerError(
                "An instance of FPS is already running. "
                "Use get_instance() to retrieve it."
            )

        if cls not in cls._instance or kwargs != {}:

            if cls not in cls._instance:
                new_obj = super().__new__(cls)
                cls._instance[cls] = new_obj
            else:
                new_obj = cls._instance[cls]

            dict.__init__(new_obj, {})
            new_obj.initialised = False

            return new_obj

        raise RuntimeError("Returning None in BaseFPS.__new__. This should not happen.")

    @classmethod
    def get_instance(cls: Type[FPS_T], *args, **kwargs) -> FPS_T:
        """Returns the running instance."""

        if cls not in cls._instance or kwargs:
            return cls(*args, **kwargs)

        return cls._instance[cls]

    @property
    def positioners(self):
        """Dictionary of positioner associated with this FPS.

        This is just a wrapper around the `.BaseFPS` instance which is a
        dictionary itself. May be deprecated in the future.

        """

        return self

    def add_positioner(
        self,
        positioner: int | Positioner,
        centre=(None, None),
    ) -> Positioner:
        """Adds a new positioner to the list, and checks for duplicates."""

        if isinstance(positioner, self.positioner_class):
            positioner_id = positioner.positioner_id
        elif isinstance(positioner, int):
            positioner_id = positioner
            positioner = self.positioner_class(positioner_id, None, centre=centre)
        else:
            raise TypeError(f"Invalid parameter positioner of type {type(positioner)}")

        if positioner_id in self.positioners:
            raise JaegerError(
                f"{self.__class__.__name__}: there is already a "
                f"positioner in the list with positioner_id "
                f"{positioner_id}."
            )

        self.positioners[positioner_id] = positioner

        return positioner


T = TypeVar("T", bound="FPS")


@dataclass
class FPS(BaseFPS["FPS"]):
    """A class describing the Focal Plane System.

    The recommended way to instantiate a new `.FPS` object is to use the `.create`
    classmethod ::

        fps = await FPS.create(...)

    which is equivalent to ::

        fps = FPS(...)
        await fps.initialise()

    Parameters
    ----------
    can
        A `.JaegerCAN` instance to use to communicate with the CAN network, or the CAN
        profile from the configuration to use, or `None` to use the default one.
    ieb
        If `True` or `None`, connects the Instrument Electronics Box PLC controller
        using the path to the IEB configuration file stored in jaeger's configuration.
        Can also be an `.IEB` instance, the path to a custom configuration file used
        to load one, or a dictionary with the configuration itself.

    Examples
    --------
    After instantiating a new `.FPS` object it is necessary to call
    `~.FPS.initialise` to retrieve the positioner layout and the status of
    the connected positioners. Note that `~.FPS.initialise` is a coroutine
    which needs to be awaited ::

        >>> fps = FPS(can='default')
        >>> await fps.initialise()
        >>> fps.positioners[4].status
        <Positioner (id=4, status='SYSTEM_INITIALIZED|
        DISPLACEMENT_COMPLETED|ALPHA_DISPLACEMENT_COMPLETED|
        BETA_DISPLACEMENT_COMPLETED')>

    """

    # __instance: ClassVar[dict[Type[FPS], FPS]] = {}

    can: JaegerCAN | str | None = None
    ieb: Union[bool, IEB, dict, str, pathlib.Path, None] = None
    status: FPSStatus = FPSStatus.IDLE | FPSStatus.TEMPERATURE_NORMAL

    def __post_init__(self):

        super().__init__()

        # Start file logger
        start_file_loggers(start_log=True, start_can=False)

        if config.CONFIG_FILE:
            log.debug(f"Using configuration from {config.CONFIG_FILE}")
        else:
            warnings.warn("Unknown configuration file.", JaegerUserWarning)

        self.loop = asyncio.get_event_loop()
        self.loop.set_exception_handler(log.asyncio_exception_handler)

        # The mapping between positioners and buses.
        self.positioner_to_bus: Dict[int, Tuple[BusABC, int | None]] = {}

        self.pid_lock: LockFile | None = None

        self._locked = False
        self.locked_by: List[int] = []

        self.observatory = config["observatory"]

        self.disabled: set[int] = set([])

        if self.ieb is None or self.ieb is True:
            self.ieb = config["ieb"]["config"]

        if isinstance(self.ieb, pathlib.Path):
            self.ieb = str(self.ieb)

        if isinstance(self.ieb, (str, dict)):
            if isinstance(self.ieb, str):
                self.ieb = os.path.expanduser(os.path.expandvars(str(self.ieb)))
                if not os.path.isabs(self.ieb):
                    self.ieb = os.path.join(os.path.dirname(__file__), self.ieb)
            try:
                self.ieb = IEB.from_config(self.ieb)
            except FileNotFoundError:
                warnings.warn(
                    f"IEB configuration file {self.ieb} cannot be loaded.",
                    JaegerUserWarning,
                )
                self.ieb = None
        elif self.ieb is False:
            self.ieb = None
        else:
            raise ValueError(f"Invalid input value for ieb {self.ieb!r}.")

        assert isinstance(self.ieb, IEB) or self.ieb is None

        self.__status_event = asyncio.Event()
        self.__temperature_task: asyncio.Task | None = None

        self._configuration: BaseConfiguration | None = None
        self._previous_configurations: list[BaseConfiguration] = []
        self._preloaded_configuration: BaseConfiguration | None = None

        # Position and status pollers
        self.pollers = PollerList(
            [
                Poller(
                    "status",
                    self.update_status,
                    delay=config["fps"]["status_poller_delay"],
                    loop=self.loop,
                ),
                Poller(
                    "position",
                    self.update_position,
                    delay=config["fps"]["position_poller_delay"],
                    loop=self.loop,
                ),
            ]
        )

    @classmethod
    async def create(
        cls,
        can=None,
        ieb=None,
        initialise=True,
        start_pollers: bool | None = None,
        enable_low_temperature: bool = True,
    ) -> "FPS":
        """Starts the CAN bus and .

        Parameters
        ----------
        initialise
            Whether to initialise the FPS.
        start_pollers
            Whether to initialise the pollers.
        kwargs
            Parameters to pass to `.FPS`.

        """

        instance = cls(can=can, ieb=ieb)
        await instance.start_can()

        if initialise:
            await instance.initialise(
                start_pollers=start_pollers,
                enable_low_temperature=enable_low_temperature,
            )

        return instance

    async def start_can(self):
        """Starts the JaegerCAN interface."""

        use_lock = config["fps"]["use_lock"]

        if use_lock and self.pid_lock is None:
            try:
                if not os.path.exists(os.path.dirname(LOCK_FILE)):
                    os.makedirs(os.path.dirname(LOCK_FILE))
                self.pid_lock = LockFile(LOCK_FILE)
            except Exception:
                raise JaegerError(
                    f"Failed creating lock file {LOCK_FILE}. "
                    "Probably another instance is running. "
                    "If that is not the case, remove the lock file and retry."
                )

        if isinstance(self.can, JaegerCAN):
            await self.can.start()
            return

        self.can = await JaegerCAN.create(self.can, fps=self)
        return True

    def add_positioner(
        self,
        positioner: int | Positioner,
        centre=(None, None),
        interface: Optional[BusABC | int] = None,
        bus: Optional[int] = None,
    ) -> Positioner:

        positioner = super().add_positioner(positioner, centre=centre)
        positioner.fps = self

        if interface is not None:
            if isinstance(interface, int):
                assert isinstance(self.can, JaegerCAN), "JaegerCAN not initialised."
                interface = self.can.interfaces[interface]
                assert isinstance(interface, BusABC), f"Invalid interface {interface!r}"

            self.positioner_to_bus[positioner.positioner_id] = (interface, bus)

        return positioner

    @property
    def configuration(self):
        """Returns the configuration."""

        return self._configuration

    @configuration.setter
    def configuration(self, new: BaseConfiguration | None):
        """Sets the new configuration."""

        # Store current configuration.
        if self._configuration is not None:
            self._previous_configurations.append(self._configuration)

        # Keep only 10 previous configurations.
        self._previous_configurations = self._previous_configurations[-10:]

        self._configuration = new

    async def initialise(
        self: T,
        start_pollers: bool | None = None,
        enable_low_temperature: bool = True,
        keep_disabled: bool = True,
        skip_fibre_assignments_check: bool = False,
    ) -> T:
        """Initialises all positioners with status and firmware version.

        Parameters
        ----------
        start_pollers
            Whether to initialise the pollers.
        enable_low_temperature
            Enables the low temperature warnings.
        keep_disabled
            Maintain the list of disabled/offline robots.
        skip_fibre_assignments_check
            Do not check fibre assignments.

        """

        if start_pollers is None:
            start_pollers = config["fps"]["start_pollers"]
        assert isinstance(start_pollers, bool)

        self.disabled = set([])
        if keep_disabled:
            for positioner in self.positioners.values():
                if positioner.offline or positioner.disabled:
                    self.disabled.add(positioner.positioner_id)

        # Clear all robots
        self.clear()
        self.positioner_to_bus = {}

        # Stop pollers while initialising
        if self.pollers.running:
            await self.pollers.stop()

        # Make sure CAN buses are connected.
        await self.start_can()

        # Test IEB connection.
        if isinstance(self.ieb, IEB):
            try:
                async with self.ieb:
                    pass
            except BaseException as err:
                warnings.warn(str(err), JaegerUserWarning)

        assert isinstance(self.can, JaegerCAN), "CAN connection not established."

        if len(self.can.interfaces) == 0:
            warnings.warn("CAN interfaces not found.", JaegerUserWarning)
            return self

        # Get the positioner-to-bus map
        await self._get_positioner_bus_map()

        # Stop poller in case they are running
        await self.pollers.stop()

        get_fw_command = self.send_command(
            CommandID.GET_FIRMWARE_VERSION,
            positioner_ids=0,
            timeout=config["fps"]["initialise_timeouts"],
        )

        assert isinstance(get_fw_command, GetFirmwareVersion)
        await get_fw_command

        if get_fw_command.status.failed:
            raise JaegerError("Failed retrieving firmware version.")

        # Loops over each reply and set the positioner status to OK. If the
        # positioner was not in the list, adds it.
        for reply in get_fw_command.replies:
            if reply.positioner_id not in self.positioners:

                if hasattr(reply.message, "interface"):
                    interface = reply.message.interface
                    bus = reply.message.bus
                else:
                    interface = bus = None

                self.add_positioner(reply.positioner_id, interface=interface, bus=bus)

            positioner = self.positioners[reply.positioner_id]
            positioner.fps = self
            positioner.firmware = get_fw_command.get_firmware()[reply.positioner_id]

            if (
                positioner.positioner_id in config["fps"]["disabled_positioners"]
                or positioner.positioner_id in self.disabled
            ):
                positioner.disabled = True
                self.disabled.add(positioner.positioner_id)

        # Add offline robots. Offline positioners are physically in the array but
        # they don't reply to commands and we need to specify their position. Once
        # That's done they behave as normal disabled robots.
        if config["fps"]["offline_positioners"] is not None:
            for pid in config["fps"]["offline_positioners"]:
                off_alpha, off_beta = config["fps"]["offline_positioners"][pid]
                if pid not in self.positioners:
                    positioner = self.add_positioner(pid)
                else:
                    positioner = self.positioners[pid]
                positioner.disabled = True
                positioner.offline = True
                positioner.alpha = off_alpha
                positioner.beta = off_beta

                self.disabled.add(positioner.positioner_id)

        # Mark as initialised here although we have some more work to do.
        self.initialised = True

        pids = sorted(list(self.keys()))
        if len(pids) > 0:
            log.info(f"Found {len(pids)} connected positioners: {pids!r}.")
        else:
            warnings.warn("No positioners found.", JaegerUserWarning)
            return self

        if len(set([pos.firmware for pos in self.values() if not pos.offline])) > 1:
            warnings.warn(
                "Found positioners with different firmware versions.",
                JaegerUserWarning,
            )

        # Stop all positioners just in case. This won't clear collided flags.
        if not self.is_bootloader():
            await self.stop_trajectory()

        # Initialise positioners
        try:
            disable_precise_moves = config["positioner"]["disable_precise_moves"]
            # if disable_precise_moves:
            #     warnings.warn("Disabling precise moves.", JaegerUserWarning)
            pos_initialise = [
                positioner.initialise(disable_precise_moves=disable_precise_moves)
                for positioner in self.values()
                if positioner.offline is False
            ]
            await asyncio.gather(*pos_initialise)
        except (JaegerError, PositionerError) as err:
            raise JaegerError(f"Some positioners failed to initialise: {err}")

        if disable_precise_moves is True and any(
            [self[i].precise_moves for i in self if self[i].offline is False]
        ):
            log.error("Unable to disable precise moves for some positioners.")

        n_non_initialised = len(
            [
                positioner
                for positioner in self.positioners.values()
                if (
                    positioner.offline is False
                    and (
                        positioner.status == positioner.flags.UNKNOWN
                        or not positioner.initialised
                    )
                )
            ]
        )

        if n_non_initialised > 0:
            raise JaegerError(f"{n_non_initialised} positioners failed to initialise.")

        if self.is_bootloader():
            bootlist = [p.positioner_id for p in self.values() if p.is_bootloader()]
            warnings.warn(
                f"Positioners in booloader mode: {bootlist!r}.",
                JaegerUserWarning,
            )
            return self

        # Check if any of the positioners are collided and if so lock the FPS.
        locked_by = []
        for positioner in self.values():
            if positioner.collision:
                locked_by.append(positioner.positioner_id)

        if len(locked_by) > 0:
            await self.lock(by=locked_by, do_warn=False, snapshot=False)
            warnings.warn(
                "The FPS was collided and has been locked.",
                JaegerUserWarning,
            )

        if config.get("safe_mode", False) is not False:
            min_beta = MIN_BETA
            if isinstance(config["safe_mode"], dict):
                min_beta = config["safe_mode"].get("min_beta", MIN_BETA)
            warnings.warn(
                f"Safe mode enabled. Minimum beta is {min_beta} degrees.",
                JaegerUserWarning,
            )

        # Disable collision detection for listed robots.
        disable_collision = config["fps"]["disable_collision_detection_positioners"]
        if len(disable_collision) > 0:
            if self.locked:
                warnings.warn(
                    "The FPS is locked. Cannot disable collision detection",
                    JaegerUserWarning,
                )

            else:
                warnings.warn(
                    "Disabling collision detection for positioners: "
                    f"{disable_collision}.",
                    JaegerUserWarning,
                )
                await self.send_command(
                    CommandID.ALPHA_CLOSED_LOOP_WITHOUT_COLLISION_DETECTION,
                    positioner_ids=disable_collision,
                )
                await self.send_command(
                    CommandID.BETA_CLOSED_LOOP_WITHOUT_COLLISION_DETECTION,
                    positioner_ids=disable_collision,
                )

        # Set robots to open loop mode
        open_loop_positioners = config["fps"].get("open_loop_positioners", [])
        if len(open_loop_positioners) > 0:
            if self.locked:
                warnings.warn(
                    "The FPS is locked. Cannot set open loop mode.",
                    JaegerUserWarning,
                )

            else:
                warnings.warn(
                    "Setting open loop mode for positioners: "
                    f"{open_loop_positioners}.",
                    JaegerUserWarning,
                )
                await self.send_command(
                    CommandID.ALPHA_OPEN_LOOP_WITHOUT_COLLISION_DETECTION,
                    positioner_ids=open_loop_positioners,
                )
                await self.send_command(
                    CommandID.BETA_OPEN_LOOP_WITHOUT_COLLISION_DETECTION,
                    positioner_ids=open_loop_positioners,
                )

        # Ensure closed loop mode for remaining robots. This does not work if
        # any of the robots is collided.
        if not self.locked:

            closed_loop_positioners = list(
                set([pid for pid in self.positioners if not self[pid].disabled])
                - set(disable_collision)
                - set(open_loop_positioners)
            )
            await self.send_command(
                CommandID.ALPHA_CLOSED_LOOP_COLLISION_DETECTION,
                positioner_ids=closed_loop_positioners,
            )
            await self.send_command(
                CommandID.BETA_CLOSED_LOOP_COLLISION_DETECTION,
                positioner_ids=closed_loop_positioners,
            )

        # Check that all the robots match the fibre assignments.
        self._check_fibre_assignments(raise_error=not skip_fibre_assignments_check)

        # Start temperature watcher.
        if self.__temperature_task is not None:
            self.__temperature_task.cancel()
        if (
            isinstance(self.ieb, IEB)
            and not self.ieb.disabled
            and enable_low_temperature
        ):
            self.__temperature_task = asyncio.create_task(self._handle_temperature())
        else:
            self.set_status(
                (self.status & ~FPSStatus.TEMPERATURE_NORMAL)
                | FPSStatus.TEMPERATURE_UNKNOWN
            )

        # Issue an update status to get the status set.
        await self.update_status()

        # Start the pollers
        if start_pollers and not self.is_bootloader():
            self.pollers.start()

        return self

    def _check_fibre_assignments(self, raise_error: bool = True):
        """Checks that all the expected robots are present."""

        if calibration is None:
            msg = "coordio.calibrations failed to import. Cannot check assignments."
            if raise_error:
                raise JaegerError(msg)
            else:
                warnings.warn(msg, JaegerUserWarning)
                return

        cal_obs = calibration.fiberAssignments.loc[self.observatory]
        cal_obs = cal_obs.loc[cal_obs.Device == "Positioner"]

        failed: bool = False

        for pid in list(cal_obs.positionerID):
            if pid not in self:
                warnings.warn(
                    f"Positioner {pid} is in fiberAssigments but not connected.",
                    JaegerUserWarning,
                )
                failed = True

        for pid in self:
            if pid not in list(cal_obs.positionerID):
                warnings.warn(
                    f"Positioner {pid} is connected but not in fiberAssigments.",
                    JaegerUserWarning,
                )
                failed = True

        if raise_error and failed:
            raise JaegerError("Some positioners do not match fiberAssignments.csv.")

    def set_status(self, status: FPSStatus):
        """Sets the status of the FPS."""

        if status != self.status:
            self.status = status
            if not self.__status_event.is_set():
                self.__status_event.set()

    async def async_status(self):
        """Generator that yields FPS status changes."""

        yield self.status
        while True:
            await self.__status_event.wait()
            yield self.status
            self.__status_event.clear()

    async def _get_positioner_bus_map(self):
        """Creates the positioner-to-bus map.

        Only relevant if the bus interface is multichannel/multibus.

        """

        assert isinstance(self.can, JaegerCAN), "CAN connection not established."

        if len(self.can.interfaces) == 1 and not self.can.multibus:
            return

        timeout = config["fps"]["initialise_timeouts"]
        id_cmd = self.send_command(
            CommandID.GET_ID,
            positioner_ids=[0],
            timeout=timeout,
        )
        await id_cmd

        # Parse the replies
        for reply in id_cmd.replies:
            iface = reply.message.interface
            bus = reply.message.bus
            self.positioner_to_bus[reply.positioner_id] = (iface, bus)

    @property
    def locked(self):
        """Returns `True` if the `.FPS` is locked."""

        return self._locked

    @property
    def moving(self):
        """Returns `True` if any of the positioners is moving."""

        return self.status & FPSStatus.MOVING

    def is_bootloader(self):
        """Returns `True` if any positioner is in bootloader mode."""

        return any([pos.is_bootloader() is not False for pos in self.values()])

    def send_command(
        self,
        command: str | int | CommandID | Command,
        positioner_ids: int | List[int] | None = None,
        data: Any = None,
        now: bool = False,
        **kwargs,
    ) -> Command:
        """Sends a command to the bus.

        Parameters
        ----------
        command
            The ID of the command, either as the integer value, a string,
            or the `.CommandID` flag. Alternatively, the `.Command` to send.
        positioner_ids
            The positioner IDs to command, or zero for broadcast. If `None`,
            sends the command to all FPS non-disabled positioners.
        data
            The bytes to send. See `.Command` for details on the format.
        interface
            The index in the interface list for the interface to use. Only
            relevant in case of a multibus interface. If `None`, the positioner
            to bus map will be used.
        bus
            The bus within the interface to be used. Only relevant in case of
            a multibus interface. If `None`, the positioner to bus map will
            be used.
        now
            If `True`, the command is sent to the CAN network immediately,
            skipping the command queue. No tracking is done for this command.
            It should only be used for emergency and shutdown commands.
        kwargs
            Extra arguments to be passed to the command.

        Returns
        -------
        command
            The command sent to the bus. The command needs to be awaited
            before it is considered done.

        """

        if not isinstance(self.can, JaegerCAN) or self.can._started is False:
            raise JaegerError("CAN connection not established.")

        if positioner_ids is None:
            positioner_ids = [p for p in self if not self[p].disabled]

        if not isinstance(command, Command):
            command_flag = CommandID(command)
            assert isinstance(command_flag, CommandID)
            CommandClass = command_flag.get_command_class()
            assert CommandClass, "CommandClass not defined"

            command = CommandClass(positioner_ids, data=data, **kwargs)

        assert isinstance(command, Command)

        broadcast = command.is_broadcast
        pids = command.positioner_ids

        if broadcast:
            if any([self[pos].disabled for pos in self]) and not command.safe:
                raise JaegerError("Some positioners are disabled.")
        else:
            if any([self[pid].disabled for pid in pids]) and not command.safe:
                raise JaegerError("Some commanded positioners are disabled.")

        if not broadcast and not all([p in self for p in pids]):
            raise JaegerError("Some positioners are not connected.")

        # Check if we are in bootloader mode.
        in_boot = [p for p in pids if p != 0 and self[p].is_bootloader()]
        if (broadcast and self.is_bootloader()) or (not broadcast and any(in_boot)):
            if not command.bootloader:
                raise JaegerError(
                    f"Cannot send command {command.command_id.name!r} "
                    "while in bootloader mode."
                )

        command_name = command.name
        command_uid = command.command_uid
        header = f"({command_name}, {command_uid}): "

        if self.locked:
            if command.safe:
                log.debug(f"FPS is locked but {command_name} is safe.")
            else:
                command.cancel(silent=True)
                raise FPSLockedError("FPS is locked.")

        elif command.move_command and self.moving:
            command.cancel(silent=True)
            raise JaegerError(
                "Cannot send move command while the "
                "FPS is moving. Use FPS.stop_trajectory() "
                "to stop the FPS."
            )

        if command.status.is_done:
            raise JaegerError(header + "trying to send a done command.")

        if not now:
            assert self.can.command_queue
            self.can.command_queue.put_nowait(command)
            can_log.debug(header + "added command to CAN processing queue.")
        else:
            self.can.send_messages(command)
            can_log.debug(header + "sent command to CAN immediately.")

        return command

    async def lock(
        self,
        stop_trajectories: bool = True,
        by: Optional[List[int]] = None,
        do_warn: bool = True,
        snapshot: bool = True,
    ):
        """Locks the `.FPS` and prevents commands to be sent.

        Parameters
        ----------
        stop_trajectories
            Whether to stop trajectories when locking. This will not
            clear any collided flags.

        """

        self._locked = True
        if do_warn:
            warnings.warn("Locking the FPS.", JaegerUserWarning)

        if stop_trajectories:
            await self.stop_trajectory()

        await self.update_status()

        axes = "?"

        if by and len(by) > 0:
            self.locked_by += by

            status_bits = self.positioners[by[0]].status
            if status_bits & PositionerStatus.COLLISION_ALPHA:
                axes = "alpha"
            if status_bits & PositionerStatus.COLLISION_BETA:
                if axes == "alpha":
                    axes = "both"
                else:
                    axes = "beta"

        if jaeger.actor_instance:
            jaeger.actor_instance.write(
                "e",
                {
                    "locked": True,
                    "locked_by": self.locked_by,
                    "locked_axes": axes,
                },
            )

        if snapshot:
            if self.locked_by is not None:
                highlight = self.locked_by[0]
            else:
                highlight = None

            log.debug(f"Saving snapshot with highlight {highlight} ({self.locked_by})")
            filename = await self.save_snapshot(highlight=highlight)
            warnings.warn(f"Snapshot for locked FPS: {filename}", JaegerUserWarning)

    async def unlock(self, force=False):
        """Unlocks the `.FPS` if all collisions have been resolved."""

        # Send STOP_TRAJECTORY. This clears the collided flags.
        await self.stop_trajectory(clear_flags=True)

        await self.update_status(timeout=2)

        for positioner in self.positioners.values():
            if positioner.collision:
                self._locked = True
                raise JaegerError(
                    "Cannot unlock the FPS until all the "
                    "collisions have been cleared."
                )

        self._locked = False
        self.locked_by = []

        return True

    def get_positions(self, ignore_disabled=False) -> numpy.ndarray:
        """Returns the alpha and beta positions as an array."""

        data = [
            (p.positioner_id, p.alpha, p.beta)
            for p in self.positioners.values()
            if ignore_disabled is False or p.disabled is False
        ]

        return numpy.array(data)

    def get_positions_dict(
        self,
        ignore_disabled=False,
    ) -> dict[int, tuple[float | None, float | None]]:
        """Returns the alpha and beta positions as a dictionary."""

        return {
            p.positioner_id: (p.alpha, p.beta)
            for p in self.positioners.values()
            if ignore_disabled is False or p.disabled is False
        }

    async def update_status(
        self,
        positioner_ids: Optional[int | List[int]] = None,
        timeout: float = 1,
    ) -> bool:
        """Update statuses for all positioners.

        Parameters
        ----------
        positioner_ids
            The list of positioners to update. If `None`, update all
            positioners.
        timeout
            How long to wait before timing out the command.

        """

        if len(self.positioners) == 0:
            return True

        if positioner_ids is None:
            positioner_ids = [0]
        elif not isinstance(positioner_ids, (list, tuple)):
            positioner_ids = [positioner_ids]

        if positioner_ids == [0]:
            valid = [pid for pid in self if self[pid].offline is False]
            n_positioners = len(valid) if len(valid) > 0 else None
        else:
            n_positioners = None

        await self.update_firmware_version(timeout=timeout)

        command = self.send_command(
            CommandID.GET_STATUS,
            positioner_ids=positioner_ids,
            n_positioners=n_positioners,
            timeout=timeout,
        )
        await command

        if command.status.failed:
            log.warning(f"{CommandID.GET_STATUS.name!r} failed during update status.")
            return False

        if len(command.replies) == 0:
            return True

        update_status_coros = []
        for pid, status_int in command.get_positioner_status().items():  # type: ignore

            if pid not in self:
                continue

            update_status_coros.append(self[pid].update_status(status_int))

        await asyncio.gather(*update_status_coros)

        # Set the status of the FPS based on positioner information.
        # First get the current bitmask without the status bit.
        current = self.status & ~FPSStatus.STATUS_BITS

        pbits = numpy.array([int(p.status) for p in self.values() if not p.disabled])

        coll_bits = PositionerStatus.COLLISION_ALPHA | PositionerStatus.COLLISION_BETA

        if ((pbits & coll_bits) > 0).any():
            self.set_status(current | FPSStatus.COLLIDED)

        elif ((pbits & PositionerStatus.DISPLACEMENT_COMPLETED) > 0).all():
            self.set_status(current | FPSStatus.IDLE)

        else:
            self.set_status(current | FPSStatus.MOVING)

        return True

    async def update_position(
        self,
        positioner_ids: Optional[int | List[int]] = None,
        timeout: float = 1,
    ) -> numpy.ndarray | bool:
        """Updates positions.

        Parameters
        ----------
        positioner_ids
            The list of positioners to update. If `None`, update all
            initialised positioners.
        timeout
            How long to wait before timing out the command.

        """

        if len(self.positioners) == 0:
            return True

        if positioner_ids is None:
            positioner_ids = [
                pos.positioner_id
                for pos in self.values()
                if pos.initialised and not pos.is_bootloader() and not pos.disabled
            ]
            if positioner_ids == []:
                return numpy.array([])

        command = await self.send_command(
            CommandID.GET_ACTUAL_POSITION,
            positioner_ids=positioner_ids,
            timeout=timeout,
        )

        if command.status.failed:
            log.error(f"{command.name} failed during update position.")
            return False

        update_position_commands = []
        for pid, position in command.get_positions().items():  # type: ignore
            if pid not in self:
                continue

            update_position_commands.append(self[pid].update_position(position))

        await asyncio.gather(*update_position_commands)

        return self.get_positions()

    async def update_firmware_version(
        self,
        positioner_ids: Optional[int | List[int]] = None,
        timeout: float = 2,
    ) -> bool:
        """Updates the firmware version of connected positioners.

        Parameters
        ----------
        positioner_ids
            The list of positioners to update. If `None`, update all
            positioners.
        timeout
            How long to wait before timing out the command.

        """

        if len(self.positioners) == 0:
            return True

        if positioner_ids is None:
            positioner_ids = [0]
        elif not isinstance(positioner_ids, (list, tuple)):
            positioner_ids = [positioner_ids]

        if positioner_ids == [0]:
            valid = [pid for pid in self if self[pid].offline is False]
            n_positioners = len(valid) if len(valid) > 0 else None
        else:
            n_positioners = None

        get_fw_command = self.send_command(
            CommandID.GET_FIRMWARE_VERSION,
            positioner_ids=0,
            timeout=timeout,
            n_positioners=n_positioners,
        )

        assert isinstance(get_fw_command, GetFirmwareVersion)
        await get_fw_command

        if get_fw_command.status.failed:
            log.error("Failed retrieving firmware version.")
            return False

        for reply in get_fw_command.replies:
            pid = reply.positioner_id
            if pid not in self.positioners:
                continue

            positioner = self.positioners[pid]
            positioner.firmware = get_fw_command.get_firmware()[pid]

        return True

    async def is_folded(self):
        """Returns `True` if the array if folded."""

        alphaL, betaL = config["kaiju"]["lattice_position"]

        await self.update_position()
        positions_array = self.get_positions(ignore_disabled=True)

        if len(positions_array) == 0:
            return False

        lattice: Any = numpy.array([alphaL, betaL])
        return numpy.allclose(positions_array[:, 1:] - lattice, 0, atol=1)

    async def stop_trajectory(self, clear_flags=False):
        """Stops all the positioners without clearing collided flags.

        Parameters
        ----------
        clear_flags
            If `True`, sends ``STOP_TRAJECTORY`` which clears collided
            flags. Otherwise sends ``SEND_TRAJECTORY_ABORT``.

        """

        if clear_flags is False:
            await self.send_command(
                "SEND_TRAJECTORY_ABORT",
                positioner_ids=None,
                timeout=0,
                now=True,
            )
        else:
            await self.send_command(
                "STOP_TRAJECTORY",
                positioner_ids=0,
                timeout=0,
                now=True,
            )

        # Check running command that are "move" and cancel them.
        assert isinstance(self.can, JaegerCAN)
        for command in self.can.running_commands.values():
            if command.move_command and not command.done():
                command.cancel(silent=True)

        self.can.refresh_running_commands()

    async def goto(
        self,
        new_positions: dict[int, tuple[float, float]],
        speed: Optional[float] = None,
        relative=False,
        use_sync_line: bool | None = None,
        go_cowboy: bool = False,
    ):
        """Sends a list of positioners to a given position.

        Parameters
        ----------
        new_positions
            The new positions as a dictionary of positioner ID to a tuple of new
            alpha and beta angles. Positioners not specified will be kept on the
            same positions.
        speed
            The speed to use.
        relative
            If `True`, ``alpha`` and ``beta`` are considered relative angles.
        use_sync_line
            Whether to use the SYNC line to start the trajectories.
        go_cowboy
            If set, does not create a ``kaiju``-safe trajectory. Use at your own risk.

        """

        try:
            traj = await goto(
                self,
                new_positions,
                relative=relative,
                speed=speed,
                use_sync_line=use_sync_line,
                go_cowboy=go_cowboy,
            )
        except Exception:
            raise
        finally:
            await self.update_status()
            await self.update_position()

        return traj

    async def send_trajectory(self, *args, **kwargs):
        """Sends a set of trajectories to the positioners.

        See the documentation for `.send_trajectory`.

        Returns
        -------
        trajectory
            The `.Trajectory` object.

        Raises
        ------
        TajectoryError
            You can inspect the `.Trajectory` object by capturing the error and
            accessing ``error.trajectory``.

        """

        return await send_trajectory(self, *args, **kwargs)

    def abort(self):
        """Aborts trajectories and stops positioners. Alias for `.stop_trajectory`."""

        return asyncio.create_task(self.stop_trajectory())

    async def send_to_all(self, *args, **kwargs):
        """Sends a command to all connected positioners.

        This method has been deprecated. Use `.send_command` with a list
        for ``positioner_ids`` instead.

        """

        raise JaegerError("send_to_all has been deprecated. Use send_command instead.")

    async def report_status(self) -> Dict[str, Any]:
        """Returns a dict with the position and status of each positioner."""

        status: Dict[str, Any] = {"positioners": {}}

        for positioner in self.positioners.values():

            pos_status = positioner.status
            pos_firmware = positioner.firmware
            pos_alpha = positioner.alpha
            pos_beta = positioner.beta

            status["positioners"][positioner.positioner_id] = {
                "position": [pos_alpha, pos_beta],
                "status": pos_status,
                "firmware": pos_firmware,
            }

        try:
            status["devices"] = self.can.device_status  # type: ignore
        except AttributeError:
            pass

        if not isinstance(self.ieb, IEB):
            status["ieb"] = False
        else:
            if self.ieb.disabled:
                status["ieb"] = False
            else:
                status["ieb"] = await self.ieb.get_status()

        return status

    async def save_snapshot(
        self,
        path: Optional[str | pathlib.Path] = None,
        collision_buffer: float | None = None,
        highlight: int | None = None,
        write_to_actor: bool = True,
    ) -> str | Axes:
        """Creates a plot with the current arrangement of the FPS array.

        Parameters
        ----------
        path
            The path where to save the plot. Defaults to
            ``/data/logs/jaeger/snapshots/MJD/fps_snapshot_<SEQ>.pdf``.
        collision_buffer
            The collision buffer.
        highlight
            A robot ID to highlight.

        """

        from jaeger.kaiju import get_snapshot

        if path is not None:
            path = str(path)

        else:
            mjd = int(Time.now().mjd)
            dirpath = os.path.join(config["fps"]["snapshot_path"], str(mjd))
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)

            path_pattern = dirpath + "/fps_snapshot_*.pdf"
            files = sorted(glob(path_pattern))

            if len(files) == 0:
                seq = 1
            else:
                seq = int(files[-1].split("_")[-1][0:4]) + 1

            path = path_pattern.replace("*", f"{mjd}_{seq:04d}")

        result = await get_snapshot(
            path,
            collision_buffer=collision_buffer,
            highlight=highlight,
        )

        if result is True and write_to_actor is True and jaeger.actor_instance:
            jaeger.actor_instance.write("i", {"snapshot": path})

        return path

    async def _handle_temperature(self):
        """Handle positioners in low temperature."""

        if not isinstance(self.ieb, IEB):
            log.error("Cannot handle low-temperature mode. IEB not present.")

        async def set_rpm(activate):
            if activate:
                rpm = config["low_temperature"]["rpm_cold"]
                log.warning(f"Low temperature mode. Setting RPM={rpm}.")
            else:
                rpm = config["low_temperature"]["rpm_normal"]
                log.warning(f"Disabling low temperature mode. Setting RPM={rpm}.")

            config["positioner"]["motor_speed"] = rpm

        async def set_idle_power(activate):
            if activate:
                ht = config["low_temperature"]["holding_torque_very_cold"]
                log.warning("Very low temperature mode. Setting holding torque.")
            else:
                ht = config["low_temperature"]["holding_torque_normal"]
                log.warning(
                    "Disabling very low temperature mode. Setting holding torque."
                )
            await self.send_command(
                CommandID.SET_HOLDING_CURRENT,
                alpha=ht[0],
                beta=ht[1],
            )

        sensor = config["low_temperature"]["sensor"]
        cold = config["low_temperature"]["cold_threshold"]
        very_cold = config["low_temperature"]["very_cold_threshold"]
        interval = config["low_temperature"]["interval"]

        while True:

            try:
                assert isinstance(self.ieb, IEB) and self.ieb.disabled is False
                device = self.ieb.get_device(sensor)
                temp = (await device.read())[0]

                # Get the status without the temperature bits.
                base_status = self.status & ~FPSStatus.TEMPERATURE_BITS

                if temp <= very_cold:
                    if self.status & FPSStatus.TEMPERATURE_NORMAL:
                        await set_rpm(True)
                        await set_idle_power(True)
                    elif self.status & FPSStatus.TEMPERATURE_COLD:
                        await set_idle_power(True)
                    else:
                        pass
                    self.set_status(base_status | FPSStatus.TEMPERATURE_VERY_COLD)

                elif temp <= cold:
                    if self.status & FPSStatus.TEMPERATURE_NORMAL:
                        await set_rpm(True)
                    elif self.status & FPSStatus.TEMPERATURE_COLD:
                        pass
                    else:
                        await set_idle_power(False)
                    self.set_status(base_status | FPSStatus.TEMPERATURE_COLD)

                else:
                    if self.status & FPSStatus.TEMPERATURE_NORMAL:
                        pass
                    elif self.status & FPSStatus.TEMPERATURE_COLD:
                        await set_rpm(False)
                    else:
                        await set_rpm(False)
                        await set_idle_power(False)
                    self.set_status(base_status | FPSStatus.TEMPERATURE_NORMAL)

            except BaseException as err:
                log.warning(
                    f"Cannot read device {sensor!r}. "
                    f"Low-temperature tracking temporarily disabled: {err}",
                )
                base_status = self.status & ~FPSStatus.TEMPERATURE_BITS
                self.set_status(base_status | FPSStatus.TEMPERATURE_UNKNOWN)

            finally:
                await asyncio.sleep(interval)

    async def shutdown(self):
        """Stops pollers and shuts down all remaining tasks."""

        if not self.is_bootloader:
            log.info("Stopping positioners and shutting down.")
            await self.stop_trajectory()

        log.debug("Stopping all pollers.")
        if self.pollers:
            await self.pollers.stop()

        log.debug("Cancelling all pending tasks and shutting down.")

        tasks = [
            task
            for task in asyncio.all_tasks(loop=self.loop)
            if task is not asyncio.current_task(loop=self.loop)
        ]
        list(map(lambda task: task.cancel(), tasks))

        await asyncio.gather(*tasks, return_exceptions=True)

        self.loop.stop()

        del self.__class__._instance[self.__class__]

    async def __aenter__(self):
        await self.initialise()
        return self

    async def __aexit__(self, *excinfo):
        await self.shutdown()

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self



# encoding: utf-8
# isort: skip_file

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING
import warnings

from sdsstools import get_config, get_logger, get_package_version
from sdsstools.configuration import __ENVVARS__

from .exceptions import JaegerUserWarning

if TYPE_CHECKING:
    from .actor import JaegerActor


NAME = "jaeger"

__version__ = get_package_version(path=__file__, package_name=NAME)


log = get_logger("jaeger", log_level=logging.WARNING)
can_log = get_logger("jaeger_can", log_level=logging.ERROR, capture_warnings=False)


# Start by loading the internal configuration file.
__ENVVARS__["OBSERVATORY"] = "?"  # Fallback

observatory = os.environ.get("OBSERVATORY", None)
if observatory is None:
    config = get_config(NAME)
else:
    config = get_config(NAME + f"_{observatory}")

# If we are not in debug mode, remove some possible warnings.
if config["debug"] is False:
    warnings.filterwarnings(
        "ignore",
        message=".+was never awaited.+",
        category=RuntimeWarning,
    )


warnings.simplefilter("always", category=JaegerUserWarning)


def start_file_loggers(start_log=True, start_can=True):

    if "files" in config and "log_dir" in config["files"]:
        log_dir = config["files"]["log_dir"]
    else:
        log_dir = "~/.jaeger"

    if start_log and log.fh is None:
        log.start_file_logger(os.path.join(log_dir, "jaeger.log"))

    if start_can and can_log.fh is None:
        can_log.start_file_logger(os.path.join(log_dir, "can.log"))


actor_instance: JaegerActor | None = None


from .can import *
from .exceptions import *
from .fps import *
from .ieb import *
from .maskbits import *
from .positioner import *
from .actor import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-03-24
# @Filename: talk.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from jaeger.commands import CommandID

from . import jaeger_parser


__all__ = ["talk"]


@jaeger_parser.command()
@click.argument("COMMAND_ID", nargs=1, type=int)
@click.argument("POSITIONER_ID", nargs=1, type=int)
@click.argument("PARAMS", nargs=-1)
async def talk(command, fps, command_id, positioner_id, params):
    """Send a direct command to the CAN network and show the replies."""

    CommandClass = CommandID(command_id).get_command_class()
    assert CommandClass

    can_command = CommandClass(positioner_id, *params)

    command.info(f"Running command {can_command.command_id.name}.")

    await fps.send_command(can_command)

    replies = can_command.replies

    for reply in replies:
        data = '"'
        for byte in reply.data:
            data += f"\\x{byte:02x}"
        data += '"'
        command.info(
            {
                "raw": [
                    reply.command_id.value,
                    reply.uid,
                    reply.response_code.value,
                    data,
                ]
            }
        )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-08-27
# @Filename: status.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import Dict, List, Tuple

from jaeger.commands import Command, CommandID
from jaeger.utils import bytes_to_int, int_to_bytes, motor_steps_to_angle


__all__ = [
    "GetID",
    "GetStatus",
    "GetCurrent",
    "GetActualPosition",
    "GetNumberTrajectories",
    "SetNumberTrajectories",
    "SwitchLEDOn",
    "SwitchLEDOff",
    "GetTemperature",
]


class GetID(Command):
    """Commands the positioners to reply with their positioner id."""

    command_id = CommandID.GET_ID
    broadcastable = True
    timeout = 1
    safe = True
    bootloader = True

    def get_ids(self) -> List:
        """Returns a list of positioners that replied back."""

        return [reply.positioner_id for reply in self.replies]


class GetStatus(Command):
    """Gets the status bits for the positioner."""

    command_id = CommandID.GET_STATUS
    broadcastable = True
    safe = True
    bootloader = True

    def get_replies(self) -> Dict[int, int]:
        return self.get_positioner_status()

    def get_positioner_status(self) -> Dict[int, int]:
        """Returns the status flag for each positioner that replied."""

        return {reply.positioner_id: bytes_to_int(reply.data) for reply in self.replies}


class GetActualPosition(Command):
    """Gets the current position of the alpha and beta arms."""

    command_id = CommandID.GET_ACTUAL_POSITION
    broadcastable = False
    safe = True

    def get_replies(self) -> Dict[int, Tuple[float, float]]:
        return self.get_positions()

    def get_positions(self) -> Dict[int, Tuple[float, float]]:
        """Returns the positions of alpha and beta in degrees.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.

        """

        positions = {}
        for reply in self.replies:
            pid = reply.positioner_id
            data = reply.data

            beta = bytes_to_int(data[4:], dtype="i4")
            alpha = bytes_to_int(data[0:4], dtype="i4")

            positions[pid] = motor_steps_to_angle(alpha, beta)

        return positions

    @staticmethod
    def encode(alpha, beta):
        """Returns the position as a bytearray in positioner units."""

        alpha_motor, beta_motor = motor_steps_to_angle(alpha, beta, inverse=True)

        alpha_bytes = int_to_bytes(int(alpha_motor), "i4")
        beta_bytes = int_to_bytes(int(beta_motor), "i4")

        data = alpha_bytes + beta_bytes

        return data


class GetCurrent(Command):
    """Gets the current of the alpha and beta motors."""

    command_id = CommandID.GET_CURRENT
    broadcastable = False
    safe = True

    def get_replies(self) -> Dict[int, Tuple[int, int]]:
        return self.get_currents()

    def get_currents(self) -> Dict[int, Tuple[int, int]]:
        """Returns a dictionary of current of alpha and beta for each positioner.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.
        """

        currents = {}

        for reply in self.replies:
            data = reply.data

            alpha = bytes_to_int(data[0:4], dtype="i4")
            beta = bytes_to_int(data[4:], dtype="i4")

            currents[reply.positioner_id] = (alpha, beta)

        return currents


class GetTemperature(Command):
    """Gets the temperature from the board temperature sensor, in C."""

    command_id = CommandID.GET_RAW_TEMPERATURE
    broadcastable = False
    safe = True

    def get_replies(self) -> Dict[int, int]:
        return self.get_replies()

    def get_temperatures(self) -> Dict[int, int]:
        """Returns the temperature in Celsius.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.
        """

        temperatures = {}

        for reply in self.replies:
            data = self.replies[0].data
            rawT = bytes_to_int(data, dtype="u4")  # Raw temperature
            temperatures[reply.positioner_id] = rawT

        return temperatures


class SwitchLEDOn(Command):
    """Switched the positioner LED on."""

    command_id = CommandID.SWITCH_LED_ON
    broadcastable = False
    safe = True


class SwitchLEDOff(Command):
    """Switched the positioner LED on."""

    command_id = CommandID.SWITCH_LED_ON
    broadcastable = False
    safe = True


class GetNumberTrajectories(Command):

    command_id = CommandID.GET_NUMBER_TRAJECTORIES
    broadcastable = False
    safe = True

    def get_replies(self):
        return self.get_number_trajectories()

    def get_number_trajectories(self) -> Dict[int, int | None]:
        """Returns the number of trajectories.

        Raises
        ------
        ValueError
            If no reply has been received or the data cannot be parsed.
        """

        number_trajectories = {}

        for reply in self.replies:
            data = reply.data
            if data == bytearray():
                number_trajectories[reply.positioner_id] = None
            else:
                number_trajectories[reply.positioner_id] = bytes_to_int(data)

        return number_trajectories


class SetNumberTrajectories(Command):

    command_id = CommandID.SET_NUMBER_TRAJECTORIES
    broadcastable = False
    safe = False

    def __init__(
        self,
        positioner_ids: int | List[int],
        moves: int | Dict[int, int],
        **kwargs,
    ):

        if isinstance(moves, int):
            data = int_to_bytes(moves)
        elif isinstance(moves, dict):
            data = {pos: int_to_bytes(moves[pos]) for pos in moves}
        else:
            raise TypeError("Invalid data type.")

        kwargs["data"] = data

        super().__init__(positioner_ids, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-10-07
# @Filename: positioner.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
from distutils.version import StrictVersion

from typing import List, Optional, Tuple

import jaeger
from jaeger import config, log, maskbits
from jaeger.can import JaegerCAN
from jaeger.commands import CommandID
from jaeger.commands.bootloader import GetFirmwareVersion
from jaeger.commands.status import GetActualPosition
from jaeger.exceptions import JaegerError, PositionerError
from jaeger.utils import StatusMixIn, bytes_to_int


__all__ = ["Positioner"]


class Positioner(StatusMixIn):
    r"""Represents the status and parameters of a positioner.

    Parameters
    ----------
    positioner_id
        The ID of the positioner
    fps
        The `~jaeger.fps.FPS` instance to which this positioner is linked to.
    centre
        The :math:`(x_{\rm focal}, y_{\rm focal})` coordinates of the
        central axis of the positioner.
    sextant
        The id of the sextant to which this positioner is connected.

    """

    def __init__(
        self,
        positioner_id: int,
        fps: jaeger.FPS | None = None,
        centre: Tuple[Optional[float], Optional[float]] = (None, None),
    ):

        self.fps = fps

        self.positioner_id = positioner_id

        self.centre = centre

        self.alpha = None
        self.beta = None
        self.speed = (None, None)
        self.firmware: str | None = None

        self.disabled = False
        self.offline = False
        self.precise_moves = True

        super().__init__(
            maskbit_flags=maskbits.PositionerStatus,
            initial_status=maskbits.PositionerStatus.UNKNOWN,
        )

    @property
    def position(self):
        """Returns a tuple with the ``(alpha, beta)`` position."""

        return (self.alpha, self.beta)

    @property
    def collision(self):
        """Returns `True` if the positioner is collided."""

        if not self.status:
            return False

        return self.status.collision

    @property
    def moving(self):
        """Returns `True` if the positioner is moving."""

        if self.status.DISPLACEMENT_COMPLETED not in self.status:
            return True

        return False

    @property
    def initialised(self):
        """Returns ``True`` if the system and datums have been initialised."""

        if self.offline:
            return True

        if self.status is None:
            return False

        if self.is_bootloader():
            if self.status != maskbits.BootloaderStatus.UNKNOWN:
                return True
            return False

        if (
            not self.status.initialised
            or self.flags.DATUM_ALPHA_INITIALIZED not in self.status
            or self.flags.DATUM_BETA_INITIALIZED not in self.status
        ):
            return False

        return True

    def reset(self):
        """Resets positioner values and statuses."""

        self.alpha = None
        self.beta = None
        self.status = self.flags.UNKNOWN
        self.firmware = None

    def _log(self, message, level=logging.DEBUG):
        """Logs a message."""

        log.log(level, f"Positioner {self.positioner_id}: {message}")

    def get_bus(self) -> Tuple[int, int | None]:
        """Returns the interface index and bus."""

        if self.fps is None:
            raise PositionerError("FPS is not defined.")

        if not isinstance(self.fps.can, JaegerCAN):
            raise PositionerError("CAN bus not found.")

        if not self.fps.can.multibus:
            return (0, None)

        iface, bus = self.fps.positioner_to_bus[self.positioner_id]
        iface_idx = self.fps.can.interfaces.index(iface)

        return (iface_idx, bus)

    async def send_command(self, command, error: Optional[str] = None, **kwargs):
        """Sends and awaits a command to the FPS for this positioner."""

        if not self.fps:
            raise PositionerError("FPS is not set.")

        command = await self.fps.send_command(
            command,
            positioner_ids=self.positioner_id,
            **kwargs,
        )

        if error and (command.status.failed or command.status.timed_out):
            raise PositionerError(error)

        return command

    async def update_position(
        self,
        position: Tuple[float, float] | None = None,
        timeout=1,
    ):
        """Updates the position of the alpha and beta arms."""

        if position is None:

            command = await self.send_command(
                CommandID.GET_ACTUAL_POSITION,
                timeout=timeout,
            )

            assert isinstance(command, GetActualPosition)

            if command.status.failed:
                self.alpha = self.beta = None
                raise PositionerError("failed updating position")

            try:
                position = command.get_positions()[self.positioner_id]
            except ValueError:
                raise PositionerError("cannot parse current position.")

        self.alpha, self.beta = position

    async def update_status(
        self,
        status: maskbits.PositionerStatus | int | None = None,
        timeout=1.0,
    ):
        """Updates the status of the positioner."""

        assert self.fps, "FPS is not set."

        # Need to update the firmware to make sure we get the right flags.
        if not self.firmware:
            await self.update_firmware_version()

        if not status:

            command = await self.send_command(
                CommandID.GET_STATUS,
                timeout=timeout,
                error="cannot get status.",
            )

            n_replies = len(command.replies)

            if n_replies == 1:
                status = int(bytes_to_int(command.replies[0].data))
            else:
                self.status = self.flags.UNKNOWN
                raise PositionerError(f"GET_STATUS received {n_replies} replies.")

        if not self.is_bootloader():
            self.flags = self.get_positioner_flags()
        else:
            self.flags = maskbits.BootloaderStatus

        self.status = self.flags(int(status))

        # Checks if the positioner is collided. If so, locks the FPS.
        # if not self.is_bootloader() and self.collision and not self.fps.locked:
        #     await self.fps.lock()
        #     raise PositionerError("collision detected. Locking the FPS.")

        return True

    async def wait_for_status(
        self,
        status: List[maskbits.PositionerStatus],
        delay=1,
        timeout: Optional[float] = None,
    ) -> bool:
        """Polls the status until it reaches a certain value.

        Parameters
        ----------
        status
            The status to wait for. Can be a list in which case it will wait
            until all the statuses in the list have been reached.
        delay
            Time, in seconds, to wait between position updates.
        timeout
            How many seconds to wait for the status to reach the desired value
            before aborting.

        Returns
        -------
        result
            Returns `True` if the status has been reached or `False` if the
            timeout limit was reached.

        """

        if self.is_bootloader():
            raise JaegerError("wait_for_status cannot be scheduled in bootloader mode.")

        if not isinstance(status, (list, tuple)):
            status = [status]

        async def status_waiter(wait_for_status):

            while True:
                await self.update_status()
                # Check all statuses in the list
                all_reached = True
                for ss in wait_for_status:
                    if ss not in self.status:
                        all_reached = False
                        break

                if all_reached:
                    return

                await asyncio.sleep(delay)

        wait_for_status = [self.flags(int(ss)) for ss in status]

        try:
            await asyncio.wait_for(status_waiter(wait_for_status), timeout)
        except asyncio.TimeoutError:
            return False

        return True

    async def initialise(self, disable_precise_moves=False):
        """Initialises the position watcher."""

        if self.offline is True:
            return True

        # Resets all.
        self.reset()

        await self.update_firmware_version()
        await self.update_status()

        # Exits if we are in bootloader mode.
        if self.is_bootloader():
            return True

        if not self.initialised:
            raise PositionerError("failed initialising.")

        # Update position only if it's not bootloader.
        await self.update_position()

        # Sets the default speed and disable precise moves
        if not self.disabled:
            await self.set_speed(
                alpha=config["positioner"]["motor_speed"],
                beta=config["positioner"]["motor_speed"],
            )

            if StrictVersion(self.firmware) < StrictVersion("04.01.17"):
                self._log("Disabling precise moves requires >=04.01.17", logging.DEBUG)
            else:
                await self.set_precise_move(mode=not disable_precise_moves)
        else:
            self.precise_moves = not disable_precise_moves

        self._log("initialisation complete.")

        return True

    async def update_firmware_version(self):
        """Updates the firmware version."""

        command = await self.send_command(
            CommandID.GET_FIRMWARE_VERSION,
            error="failed retrieving firmware version.",
        )

        assert isinstance(command, GetFirmwareVersion)

        firmware = command.get_firmware()[self.positioner_id]
        assert firmware is not None and isinstance(firmware, str)

        self.firmware = firmware
        self.flags = self.get_positioner_flags()

        self._log(f"firmware {self.firmware}")

        return True

    def get_positioner_flags(self):
        """Returns the correct position maskbits from the firmware version."""

        assert self.firmware, "Firmware is not set."

        if self.is_bootloader():
            return maskbits.BootloaderStatus

        if StrictVersion(self.firmware) < StrictVersion("04.01.00"):
            return maskbits.PositionerStatusV4_0
        else:
            return maskbits.PositionerStatus

    def is_bootloader(self):
        """Returns True if we are in bootloader mode."""

        if self.offline:
            return False

        if self.firmware is None:
            return None

        return self.firmware.split(".")[1] == "80"

    async def set_position(self, alpha: float, beta: float):
        """Sets the internal position of the motors."""

        set_position_command = await self.send_command(
            CommandID.SET_ACTUAL_POSITION,
            alpha=float(alpha),
            beta=float(beta),
            error="failed setting position.",
        )

        self._log(f"position set to ({alpha:.2f}, {beta:.2f})")

        return set_position_command

    async def set_speed(self, alpha: float, beta: float, force=False):
        """Sets motor speeds.

        Parameters
        ----------
        alpha
            The speed of the alpha arm, in RPM on the input.
        beta
            The speed of the beta arm, in RPM on the input.
        force
            Allows to set speed limits outside the normal range.

        """

        MIN_SPEED = 0
        MAX_SPEED = 5000

        if (
            alpha < MIN_SPEED
            or alpha > MAX_SPEED
            or beta < MIN_SPEED
            or beta > MAX_SPEED
        ) and not force:
            raise PositionerError("speed out of limits.")

        speed_command = await self.send_command(
            CommandID.SET_SPEED,
            alpha=float(alpha),
            beta=float(beta),
            error="failed setting speed.",
        )

        self.speed = (alpha, beta)

        self._log(f"speed set to ({alpha:.2f}, {beta:.2f})")

        return speed_command

    async def set_precise_move(self, mode, alpha=True, beta=True):
        """Switches the precise moves on alpha and beta."""

        cmds = []

        if not alpha and not beta:
            raise PositionerError("positioneralpha or beta need to be True.")

        if alpha:
            if mode is True:
                cmds.append(self.send_command(CommandID.SWITCH_ON_PRECISE_MOVE_ALPHA))
            else:
                cmds.append(self.send_command(CommandID.SWITCH_OFF_PRECISE_MOVE_ALPHA))

        if beta:
            if mode is True:
                cmds.append(self.send_command(CommandID.SWITCH_ON_PRECISE_MOVE_BETA))
            else:
                cmds.append(self.send_command(CommandID.SWITCH_OFF_PRECISE_MOVE_BETA))

        cmds = await asyncio.gather(*cmds)

        if any([cmd.status != maskbits.CommandStatus.DONE for cmd in cmds]):
            raise PositionerError("failed switching precise moves.")

        self.precise_moves = mode

        return True

    def _can_move(self):
        """Returns `True` if the positioner can be moved."""

        if self.moving or self.collision or not self.initialised or not self.status:
            return False

        if self.flags == maskbits.PositionerStatusV4_0:
            return True
        else:
            PS = maskbits.PositionerStatus
            invalid_bits = [
                PS.COLLISION_DETECT_ALPHA_DISABLE,
                PS.COLLISION_DETECT_BETA_DISABLE,
            ]
            for b in invalid_bits:
                if b in self.status:
                    self._log(f"canot move; found status bit {b.name}.", logging.ERROR)
                    return False
            if (
                PS.CLOSED_LOOP_ALPHA not in self.status
                or PS.CLOSED_LOOP_BETA not in self.status
            ):
                self._log(
                    "canot move; positioner not in closed loop mode.",
                    logging.ERROR,
                )
                return False
            return True

    async def _goto_position(self, alpha: float, beta: float, relative=False):
        """Go to a position."""

        if relative:
            command_id = CommandID.GO_TO_RELATIVE_POSITION
        else:
            command_id = CommandID.GO_TO_ABSOLUTE_POSITION

        return await self.send_command(
            command_id,
            alpha=float(alpha),
            beta=float(beta),
            error="failed going to position.",
        )

    async def home(self, alpha: bool = True, beta: bool = True):
        """Homes the positioner.

        Zeroes the positioner by counter-clockwise rotating alpha and beta
        until they hit the hardstops. Blocks until the move is complete.

        Parameters
        ----------
        alpha
            Home the alpha arm.
        beta
            Home the beta arm.

        """

        if self.moving:
            raise PositionerError("positioner is already moving.")

        if not self.fps:
            raise PositionerError("the positioner is not linked to a FPS instance.")

        if alpha and beta:
            await self.send_command(
                "GO_TO_DATUMS",
                error="failed while sending GO_TO_DATUMS command.",
            )
        elif alpha:
            await self.send_command(
                "GO_TO_DATUM_ALPHA",
                error="failed while sending GO_TO_DATUMS_ALPHA command.",
            )
        elif beta:
            await self.send_command(
                "GO_TO_DATUM_BETA",
                error="failed while sending GO_TO_DATUMS_BETA command.",
            )

        self._log("waiting to home.")
        await self.wait_for_status(self.flags.DISPLACEMENT_COMPLETED)

        self._log("homed.", logging.INFO)

    async def set_loop(self, motor="both", loop="closed", collisions=True):
        """Sets the control loop for a motor.

        These parameters are cleared after a restart. The motors revert to
        closed loop with collision detection.

        Parameters
        ----------
        motor
            The motor to which these changes apply, either ``'alpha`'``,
            ``'beta'``, or ``'both'``.
        loop
            The type of control loop, either ``'open'`` or ``'closed'``.
        collisions
            Whether the firmware should automatically detect collisions and
            stop the positioner.

        """

        if motor == "both":
            motors = ["alpha", "beta"]
        else:
            motors = [motor]

        for motor in motors:

            command_name = motor.upper() + "_" + loop.upper() + "_LOOP"
            if collisions:
                command_name += "_COLLISION_DETECTION"
            else:
                command_name += "_WITHOUT_COLLISION_DETECTION"

            await self.send_command(
                command_name, error=f"failed setting loop for {motor}."
            )

            self._log(
                f"set motor={motor!r}, loop={loop!r}, "
                f"detect_collision={collisions}",
            )

        return True

    def __repr__(self):

        return (
            f"<Positioner (id={self.positioner_id}, "
            f"status={self.status!s}, initialised={self.initialised})>"
        )

    async def get_number_trajectories(self) -> int | None:
        """Returns the number of trajectories executed by the positioner.

        Will return `None` if the firmware does not support the
        ``GET_NUMBER_TRAJECTORIES``.

        """

        if not self.firmware:
            return None

        if StrictVersion(self.firmware) < StrictVersion("04.01.21"):
            return None

        assert self.fps

        cmd = await self.fps.send_command(
            CommandID.GET_NUMBER_TRAJECTORIES,
            positioner_ids=[self.positioner_id],
            timeout=1,
        )

        n_traj = cmd.get_replies()[self.positioner_id]

        return n_traj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-12
# @Filename: ieb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Tuple

import click

from clu.parsers.click import pass_args

from jaeger.fps import FPS
from jaeger.ieb import IEB, _get_category_data

from . import jaeger_parser


if TYPE_CHECKING:
    from clu.command import Command

    from .. import JaegerActor


__all__ = ["ieb"]


@jaeger_parser.group()
@pass_args()
@click.pass_context
def ieb(ctx, command, fps):
    """Manages the IEB."""

    ieb = fps.ieb

    if not ieb or (ieb.disabled and ctx.invoked_subcommand != "enable"):
        command.fail(error="ieb not connected.")
        raise click.Abort()

    return


@ieb.command()
async def enable(command, fps):
    """Re-enables the IEB."""

    if fps.ieb is None or not isinstance(fps.ieb, IEB):
        return command.fail("IEB object does not exist.")

    fps.ieb.enable()

    return command.finish()


@ieb.command()
async def disable(command, fps):
    """Disables the IEB."""

    if fps.ieb is None or not isinstance(fps.ieb, IEB):
        return command.fail("IEB object does not exist.")

    fps.ieb.disabled = True

    return command.finish()


@ieb.command()
async def status(command, fps):
    """Outputs the status of the devices."""

    categories = sorted(fps.ieb.get_categories())

    status_data = {}
    for category in categories:
        if category not in command.actor.model.schema["properties"]:
            command.warning(f"Unknown device category {category!r}.")
        status_data[category] = await _get_category_data(command, category)

    return command.finish(message=status_data, concatenate=True)


@ieb.command()
@click.argument("device_names", metavar="DEVICES", type=str, nargs=-1)
@click.argument("VALUE", type=click.FloatRange(0, 100))
async def fbi(command, fps: FPS, device_names: Tuple[str], value: float):
    """Control the power output (0-100%) of the fibre back illuminator (FBI)."""

    raw_value = 32 * int(1023 * (value / 100))

    if not isinstance(fps.ieb, IEB) or fps.ieb.disabled:
        return command.fail(error="IEB is not conencted or is disabled.")

    if len(device_names) == 0:
        return command.fail(error="No devices provided.")

    for device_name in device_names:
        try:
            device = fps.ieb.get_device(device_name)
        except ValueError:
            return command.fail(error=f"Cannot find device {device_name!r}.")

        if device.mode != "holding_register":
            return command.fail(
                error=f"Invalid device mode for {device_name!r}: {device.__type__}."
            )

        await device.write(raw_value)

    await asyncio.sleep(0.2)
    fbi_led = await _get_category_data(command, "fbi_led")

    return command.finish(fbi_led=fbi_led)


@ieb.command()
@click.option("-v", "--verbose", is_flag=True, help="Shows extra information.")
async def info(command: Command[JaegerActor], fps: FPS, verbose=False):
    """Shows information about the IEB layout in a human-readable format."""

    ieb = fps.ieb

    if not isinstance(ieb, IEB):
        return command.fail(error="IEB is not conencted or is disabled.")

    modules = sorted(ieb.modules.keys())
    categories = set()

    command.info(text="Modules:")

    for module_name in modules:
        module = ieb.modules[module_name]

        command.info(text=f"  {module.name}:")

        if module.description != "":
            command.info(text=f"    description: {module.description}")

        if verbose:
            if module.model:
                command.info(text=f"    model: {module.model}")
            if module.mode:
                command.info(text=f"    mode: {module.mode}")
            if module.channels:
                command.info(text=f"    channels: {module.channels}")

        command.info(text="    Devices:")

        for dev in ieb[module_name].devices.values():
            command.info(text=f"      {dev.name}:")
            if dev.category != "":
                command.info(text=f"        category: {dev.category}")
                categories.add(dev.category)
            if dev.description != "":
                command.info(text=f"        description: {dev.description}")

            if verbose:
                if dev.address:
                    command.info(text=f"        address: {dev.address}")
                if dev.channel:
                    command.info(text=f"        channel: {dev.channel}")
                if dev.__type__:
                    command.info(text=f"        type: {dev.__type__}")
                    if dev.__type__ == "relay" and dev.relay_type:
                        command.info(text=f"        relay type: {dev.relay_type}")

    if len(categories) > 0 and command.actor and command.actor.model:

        command.info(text="")
        command.info(text="Keywords:")

        for category in sorted(categories):
            cat_data = command.actor.model.schema["properties"][category]
            items = [item["title"] for item in cat_data["items"]]
            command.info(text=f"  {category}=[{', '.join(items)}]")

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-27
# @Filename: chiller.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

import click

from jaeger.ieb import Chiller

from . import JaegerCommandType, command_parser


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = ["chiller"]


@command_parser.group()
def chiller(*args):
    """Shows the status and controlls the chiller."""

    pass


@chiller.command()
async def status(command: JaegerCommandType, fps: FPS):
    """Shows the status of the chiller."""

    chiller = Chiller.create()
    if chiller is None:
        return command.fail("Cannot access the chiller.")

    names = [
        ("chiller_temperature_value", "DISPLAY_VALUE"),
        ("chiller_temperature_setpoint", "TEMPERATURE_USER_SETPOINT"),
        ("chiller_flow_value", "STATUS_FLUID_FLOW"),
        ("chiller_flow_setpoint", "FLOW_USER_SETPOINT"),
    ]

    keywords = {}

    for _ in range(10):  # Try ten times or fail.
        try:
            for key, dev_name in names:
                value = round((await chiller.read_device(dev_name))[0], 1)
                keywords[key] = value
        except Exception:
            await asyncio.sleep(1)
            continue

        # Broadcast so that we can call it from chiller set.
        command.info(message=keywords, broadcast=True)

        return command.finish()

    return command.fail("Timed out getting chiller status.")


@chiller.command()
async def disable(command: JaegerCommandType, fps: FPS):
    """Disables the chiller watcher."""

    if not command.actor.chiller:
        return command.fail("The chiller bot is not running.")

    await command.actor.chiller.stop()

    return command.finish("Chiller watcher has been disabled.")


@chiller.command()
@click.argument("MODE", type=click.Choice(["temperature", "flow"]))
@click.argument("VALUE", type=str)
async def set(command: JaegerCommandType, fps: FPS, mode: str, value: str | float):
    """Shows the temperature or flow of the chiller.

    The value of the chiller temperature or flow can be set to auto, disable,
    or a fixed numerical value.

    """

    actor = command.actor
    assert actor

    if isinstance(value, str) and value.lower() not in ["disable", "auto"]:
        try:
            value = float(value)
        except ValueError:
            return command.fail(f"Invalid value {value!r}.")

        if mode == "temperature" and value < 0.1:
            return command.fail("Minimum temperature is 0.1 C.")

        if mode == "flow" and (value < 0.1 or value > 15):
            return command.fail("Invalid flow rate.")

    if command.actor.chiller is None:
        return command.fail("The chiller bot does not exist.")

    if value == "auto":
        setattr(command.actor.chiller, mode, True)
        command.info(f"Chiller {mode} set to auto.")
    elif value == "disable":
        setattr(command.actor.chiller, mode, False)
        command.info(f"Chiller {mode} tracking disabled.")
    else:
        setattr(command.actor.chiller, mode, value)

    await command.actor.chiller.restart()
    await asyncio.sleep(3)
    await command.send_command("jaeger", "chiller status")

    return command.finish()
